from requests import Session
from tableau_utilities.tableau_server.base import Base


class Create(Base):
    """ Core Create functionality of the TableauServer class """
    def __init__(self, parent):
        super().__init__(parent)

    def project(self, name, description='', content_permissions='LockedToProject'):
        """ Creates a project.

        Args:
            name (str): The name of the project
            description (str): The description of the project
            content_permissions (str): The content permissions, e.g. LockedToProject
        """
        self._post(
            f'{self.url}/projects',
            {
                'project': {
                    'name': name,
                    'description': description,
                    'contentPermissions': content_permissions
                }
            }
        )

    def group(self, name, minimum_site_role='Viewer'):
        """ Creates a group.

        Args:
            name (str): The name of the Group
            minimum_site_role (str): The minimum site role of the group, e.g. Viewer
        """
        self._post(
            f'{self.url}/groups',
            {
                'group': {
                    'name': name,
                    'minimumSiteRole': minimum_site_role
                }
            }
        )



import tableau_utilities.tableau_server.tableau_server_objects as tso
from requests import Session
from tableau_utilities.tableau_server.static import transform_tableau_object
from tableau_utilities.tableau_server.base import Base


class Update(Base):
    """ Core Update functionality of the TableauServer class """
    def __init__(self, parent):
        super().__init__(parent)

    def datasource_connection(self, datasource_id, connection: tso.Connection):
        conn_dict = connection.dict()
        if 'user_name' in conn_dict:
            conn_dict.setdefault('userName', conn_dict.pop('user_name'))
        if 'embed_password' in conn_dict:
            conn_dict.setdefault('embedPassword', conn_dict.pop('embed_password'))
        if 'password' in conn_dict:
            conn_dict.setdefault('password', conn_dict.pop('password'))
        if 'server_address' in conn_dict:
            conn_dict.setdefault('serverAddress', conn_dict.pop('server_address'))
        if 'query_tagging_enabled' in conn_dict:
            conn_dict.setdefault('queryTaggingEnabled', conn_dict.pop('query_tagging_enabled'))

        content = self._put(
            f'{self.url}/datasources/{datasource_id}/connections/{connection.id}',
            json={'connection': conn_dict}
        )

        transform_tableau_object(content['connection'])
        return tso.Connection(**content['connection'])



import logging
import sys
import re
import xml.etree.ElementTree as ET
import xmltodict
from datetime import datetime
from dataclasses import dataclass, astuple
from typing import Literal
from tableau_utilities.general.funcs import transform_tableau_object

@dataclass
class TableauFileObject:
    """
        A Tableau File object is any element in a Tableau Online/Server object that can be downloaded,
        i.e. Columns, Folders, Connections, etc, in Datasources/Workbooks.

        This is the base class for a Tableau File object,
        with the minimum required attributes and functionality.
        Child classes will inherit the functionality provided by this base class.

        The data types of attributes from child class will be converted to the appropriate type,
        if it is provided as a string instead.
    """

    def __bool__(self):
        return len([v for k, v in self.__dict__.items() if v and k != 'tag']) > 0

    def __existing_str_attr(self, attr: str):
        """ Returns: True if the attribute exists and is a string """
        return (
            hasattr(self, attr)
            and getattr(self, attr)
            and isinstance(getattr(self, attr), str)
        )

    def __to_int(self, attr: str):
        """
            Set the attribute to an integer,
            if the class has the attribute,
            and if the attribute is a string.
        """
        if self.__existing_str_attr(attr):
            setattr(self, attr, int(getattr(self, attr)))

    def __to_bool(self, attr: str):
        """
            Set the attribute to a boolean,
            if the class has the attribute,
            and if the attribute is a string.
        """
        if self.__existing_str_attr(attr):
            setattr(self, attr, getattr(self, attr).lower() in ['true', 'yes'])

    def __to_datetime(self, attr: str):
        """
            Set the attribute to a datetime,
            if the class has the attribute,
            and if the attribute is a string.
        """
        if self.__existing_str_attr(attr):
            setattr(self, attr, datetime.strptime(getattr(self, attr), '%Y-%m-%d %H:%M:%S.%f'))

    def __post_init__(self):
        # Convert to Boolean
        self.__to_bool('contains_null')
        self.__to_bool('datatype_customized')
        self.__to_bool('extract_engine')
        self.__to_bool('enabled')
        self.__to_bool('hidden')
        self.__to_bool('incremental_updates')
        self.__to_bool('user_specific')
        # Convert to Integer
        self.__to_int('approx_count')
        self.__to_int('collation_flag')
        self.__to_int('count')
        self.__to_int('ordinal')
        self.__to_int('port')
        self.__to_int('precision')
        self.__to_int('scale')
        self.__to_int('width')
        # Convert to Datetime
        self.__to_datetime('timestamp_start')

    def dict(self):
        pass

    def xml(self):
        """ Returns the FileObject as an XML Element """
        return ET.fromstring(xmltodict.unparse({self.tag: self.dict()}, pretty=True))


class TableauFileObjects(list):
    """
        A list of FileObject items from a Tableau file,
        i.e. Columns, Folders, Connections, etc
    """
    def __init__(self, seq=(), item_class=TableauFileObject, tag=''):
        # Set and validate
        self._item_class = item_class
        self.tag = tag
        # Enforce listed items
        if isinstance(seq, (dict, TableauFileObject)):
            seq = [seq]
        super().__init__(self.__validate_item(item) for item in seq)

    def __getitem__(self, item):
        if isinstance(item, int):
            return super().__getitem__(item)
        return super().__getitem__(self.index(item))

    def __setitem__(self, item, newitem):
        if isinstance(item, int):
            super().__setitem__(item, newitem)
        else:
            super().__setitem__(self.index(item), newitem)

    def __eq__(self, other):
        same = True
        for item in self:
            if item not in other:
                same = False
        for item in other:
            if item not in self:
                same = False
        return same

    def __del__(self):
        self.clear()

    def __validate_item(self, item):
        if isinstance(item, dict):
            _item = transform_tableau_object(item)
            try:
                return self._item_class(**_item)
            except TypeError as err:
                print(item)
                print(_item)
                raise TypeError(err) from err
        elif not isinstance(item, self._item_class):
            raise TypeError(f'Item must be of type {self._item_class.__name__} or dict, not {item.__class__.__name__}')
        return item

    def _to_dict(self):
        """ Converts all items into dicts """
        for idx, item in enumerate(self):
            if isinstance(item, self._item_class):
                self[idx] = item.dict()

    def _to_obj(self):
        """ Converts all items into the appropriate Tableau FileObject """
        for idx, item in enumerate(self):
            if isinstance(item, dict):
                item = transform_tableau_object(item)
                self[idx] = self._item_class(**item)

    def add(self, item):
        """ Add the provided item to the list

        Args:
            item (dict|TableauFileObject): The item to be added
        """
        item = self.__validate_item(item)
        if item not in self:
            self.append(item)

    def extend(self, items):
        """ Extends the Datasource items with another list of Datasource items.

        Args:
            items (TableauFileObjects[TableauFileObject]|list[TableauFileObject]): A list of Tableau FileObjects
        """
        for item in items:
            self.add(item)

    def update(self, item):
        """ Add the provided item to the list

        Args:
            item (dict|TableauFileObject): The item to be added
        """
        item = self.__validate_item(item)
        self[item] = item

    def get(self, item):
        """ Get the item from the list

        Args:
            item (dict|str|TableauFileObject): The item to get

        Returns: The Tableau FileObject
        """
        if isinstance(item, int):
            return self[item]
        for _item in self:
            if _item == item:
                return _item

    def delete(self, item):
        """ Delete the item from the list

        Args:
            item (dict|str|TableauFileObject): The item to delete
        """
        self.remove(item)

    def pop(self, item):
        """ Delete and return the item from the list

        Args:
            item (dict|str|TableauFileObject|int): The item to delete

        Returns: The Tableau FileObject
        """
        if isinstance(item, int):
            return super().pop(item)
        else:
            return super().pop(self.index(item))

    def xml(self):
        """ Returns the TableauFileObjects as an XML Element """
        self._to_dict()
        xml = ET.fromstring(xmltodict.unparse({self.tag: self}, pretty=True))
        self._to_obj()
        return xml


@dataclass
class Column(TableauFileObject):
    """ The Column Tableau file object """
    name: str
    datatype: Literal['boolean', 'date', 'datetime', 'integer', 'real', 'string', 'table']
    role: Literal['dimension', 'measure']
    type: Literal['nominal', 'ordinal', 'quantitative']
    tag: str = 'column'
    default_role: str = None
    default_type: str = None
    default_format: str = None
    auto_column: str = None
    xmlns: str = None
    semantic_role: str = None
    caption: str = None
    desc: str = None
    calculation: str = None
    aggregation: str = None
    user_auto_column: str = None
    param_domain_type: str = None
    value: str = None
    alias: str = None
    ordinal: int = None
    datatype_customized: bool = None
    hidden: bool = None
    aliases: list = None
    members: dict = None
    range: dict = None
    fiscal_year_start: int = None
    visual_totals: str = None
    ns0_auto_column: str = None
    xmlns_ns0: str = None

    def __post_init__(self):
        if not re.match(r'^\[.+]$', self.name):
            self.name = f'[{self.name}]'
        if self.aliases and isinstance(self.aliases, dict):
            if isinstance(self.aliases['alias'], list):
                self.aliases = self.aliases['alias']
            else:
                self.aliases = [self.aliases['alias']]
        if self.desc and isinstance(self.desc, dict):
            self.desc = self.desc['formatted-text']['run']
        if self.calculation and isinstance(self.calculation, dict):
            self.calculation = self.calculation['@formula'] if self.calculation['@class'] == 'tableau' else None
        super().__post_init__()

    def __hash__(self):
        return hash(str(astuple(self)))

    def __eq__(self, other):
        name = ''
        if isinstance(other, str):
            name = other
        elif isinstance(other, dict):
            name = other.get('name')
        elif isinstance(other, (Column, object)):
            name = other.name

        if not re.match(r'^\[.+]$', name):
            name = f'[{name}]'

        return self.name == name

    def dict(self):
        output = {
            '@name': self.name,
            '@datatype': self.datatype,
            '@role': self.role,
            '@type': self.type

        }
        if self.semantic_role is not None:
            output['@semantic-role'] = self.semantic_role
        if self.auto_column is not None:
            output['@ns0:auto-column'] = self.auto_column
        if (self.xmlns or self.xmlns_ns0) is not None:
            output['@xmlns:ns0'] = self.xmlns or self.xmlns_ns0
        if self.ns0_auto_column is not None:
            output['@ns0:auto-column'] = self.ns0_auto_column
        if self.default_format is not None:
            output['@default-format'] = self.default_format
        if self.caption is not None:
            output['@caption'] = self.caption
        if self.aggregation is not None:
            output['@aggregation'] = self.aggregation
        if self.user_auto_column is not None:
            output['@user:auto-column'] = self.user_auto_column
        if self.default_role is not None:
            output['@default-role'] = self.default_role
        if self.default_type is not None:
            output['@default-type'] = self.default_type
        if self.param_domain_type is not None:
            output['@param-domain-type'] = self.param_domain_type
        if self.alias is not None:
            output['@alias'] = self.alias
        if self.value is not None:
            output['@value'] = self.value
        if self.ordinal is not None:
            output['@ordinal'] = str(self.ordinal)
        if self.members is not None:
            output['members'] = self.members
        if self.range is not None:
            output['range'] = self.range
        if self.hidden is not None:
            output['@hidden'] = str(self.hidden).lower()
        if self.datatype_customized is not None:
            output['@datatype-customized'] = str(self.datatype_customized).lower()
        if self.calculation is not None:
            output['calculation'] = {'@class': 'tableau', '@formula': self.calculation}
        if self.desc is not None:
            output['desc'] = {'formatted-text': {'run': self.desc}}
        if self.aliases is not None:
            output['aliases'] = dict()
            output['aliases']['alias'] = self.aliases
        if self.fiscal_year_start is not None:
            output['@fiscal-year-start'] = self.fiscal_year_start
        if self.visual_totals is not None:
            output['@visual-totals'] = self.visual_totals

        return output


@dataclass
class Relation(TableauFileObject):
    """ The Relation Tableau file object """
    type: str
    tag: str = 'relation'
    name: str = None
    table: str = None
    text: str = None
    connection: str = None
    relation: TableauFileObjects = None

    def __hash__(self):
        return hash(str(astuple(self)))

    def __eq__(self, other):
        if isinstance(other, str):
            return self.name == other
        if isinstance(other, dict):
            return self.name == other.get('name')
        if isinstance(other, (FolderItem, object)):
            return self.name == other.name
        return False

    def __post_init__(self):
        if self.relation is not None:
            self.relation = TableauFileObjects(self.relation, item_class=Relation, tag=self.tag)
        else:
            self.relation = TableauFileObjects(item_class=Relation, tag=self.tag)
        super().__post_init__()

    def dict(self):
        dictionary = {'@type': self.type}
        if self.name is not None:
            dictionary['@name'] = self.name
        if self.connection is not None:
            dictionary['@connection'] = self.connection
        if self.table is not None:
            dictionary['@table'] = self.table
        if self.text is not None:
            dictionary['@text'] = self.text
        if self.relation is not None:
            dictionary['relation'] = [r.dict() for r in self.relation]
        return dictionary


@dataclass
class MappingCol(TableauFileObject):
    """ The mapping Col Tableau file object """
    key: str
    value: str
    tag: str = 'map'

    def __post_init__(self):
        if not re.match(r'^\[.+]$', self.key):
            self.key = f'[{self.key}]'
        if not re.match(r'^\[.+]$', self.value):
            table, column = self.value.split('.')
            self.value = f'[{table}].[{column}]'
        super().__post_init__()

    def __hash__(self):
        return hash(str(astuple(self)))

    def __eq__(self, other):
        key = ''
        value = None
        if isinstance(other, str):
            key = other
        elif isinstance(other, dict):
            key = other.get('key')
            value = other.get('value')
        elif isinstance(other, (MappingCol, object)):
            key = other.key
            value = other.value

        if not re.match(r'^\[.+]$', key):
            key = f'[{key}]'

        if value:
            return self.key == key and self.value == value
        else:
            return self.key == key

    def dict(self):
        return {'@key': self.key, '@value': self.value}


@dataclass
class FolderItem(TableauFileObject):
    """ The FolderItem Tableau file object, is an Item of the Folder.folder_item list """
    name: str
    type: str = 'field'
    tag: str = 'folder-item'

    def __hash__(self):
        return hash(str(astuple(self)))

    def __eq__(self, other):
        if isinstance(other, str):
            return self.name == other
        if isinstance(other, dict):
            return self.name == other.get('name') and self.type == other.get('type')
        if isinstance(other, (FolderItem, object)):
            return self.name == other.name and self.type == other.type
        return False

    def dict(self):
        return {'@name': self.name, '@type': self.type}


@dataclass
class Folder(TableauFileObject):
    """ The Folder Tableau file object """
    name: str
    tag: str = 'folder'
    role: str = None
    folder_item: TableauFileObjects[FolderItem] = None

    def __post_init__(self):
        if self.folder_item is not None:
            self.folder_item = TableauFileObjects(self.folder_item, item_class=FolderItem, tag='folder-item')
        else:
            self.folder_item = TableauFileObjects(item_class=FolderItem, tag='folder-item')
        super().__post_init__()

    def __hash__(self):
        return hash(str(astuple(self)))

    def __eq__(self, other):
        if isinstance(other, str):
            return self.name == other
        if self.role is not None:
            if isinstance(other, dict):
                return self.name == other.get('name') and self.role == other.get('role')
            if isinstance(other, (Column, object)):
                return self.name == other.name and self.role == other.role
        else:
            if isinstance(other, dict):
                return self.name == other.get('name')
            if isinstance(other, (Column, object)):
                return self.name == other.name
        return False

    def dict(self):
        output = {'@name': self.name}
        if self.role:
            output['@role'] = self.role
        if self.folder_item:
            output['folder-item'] = list()
            for folder_item in self.folder_item:
                output['folder-item'].append(folder_item.dict())
        return output


@dataclass
class FoldersCommon(TableauFileObject):
    """ The FoldersCommon Tableau file object """
    folder: TableauFileObjects[Folder] = None
    tag: str = 'folders-common'

    def __post_init__(self):
        if self.folder is not None:
            self.folder = TableauFileObjects(self.folder, item_class=Folder, tag='folder')
        else:
            self.folder = TableauFileObjects(item_class=Folder, tag='folder')
        super().__post_init__()

    def __getitem__(self, item):
        return self.folder[item]

    def __setitem__(self, key, value):
        self.folder[key] = value

    def __str__(self):
        return str(self.folder)

    def __repr__(self):
        return repr(self.folder)

    def __iter__(self):
        return iter(self.folder)

    def __eq__(self, other):
        if isinstance(other, FoldersCommon):
            other = other.folder
        for item in self.folder:
            if item not in other:
                return False
        for item in other:
            if item not in self.folder:
                return False
        return True

    def get(self, folder):
        """ Get the Folder object from FoldersCommon

        Args:
            folder (Folder|str|int): The Folder, Folder.name, or Folder index, to get

        Returns: The Folder object
        """
        return self.folder.get(folder)

    def add(self, folder):
        """ Add the Folder object to FoldersCommon

        Args:
            folder (Folder): The Folder object to add
        """
        self.folder.add(folder)

    def delete(self, folder):
        """ Delete the Folder object from FoldersCommon

        Args:
            folder (Folder|str|int): The Folder, Folder.name, or Folder index, to delete
        """
        self.folder.delete(folder)

    def update(self, folder):
        """ Update the FoldersCommon Folder object

        Args:
            folder (Folder): The Folder object to update
        """
        self.folder.update(folder)

    def dict(self):
        return {'folder': [f.dict() for f in self.folder]}


@dataclass
class DrillPath(TableauFileObject):
    """ The DrillPath Tableau file object """
    name: str
    field: list[str] = None
    tag: str = 'drill-path'

    def __hash__(self):
        return hash(str(astuple(self)))

    def __eq__(self, other):
        if isinstance(other, str):
            return self.name == other
        else:
            if isinstance(other, dict):
                return self.name == other.get('name')
        return False

    def dict(self):
        output = {'@name': self.name}
        if self.field:
            output['field'] = self.field
        return output


@dataclass
class DrillPaths(TableauFileObject):
    """ The DrillPaths Tableau file object """
    drill_path: TableauFileObjects[DrillPath] = None
    tag: str = 'drill-paths'

    def __post_init__(self):
        if self.drill_path is not None:
            self.drill_path = TableauFileObjects(self.drill_path, item_class=DrillPath, tag='drill-path')
        else:
            self.drill_path = TableauFileObjects(item_class=DrillPath, tag='drill-path')
        super().__post_init__()

    def __getitem__(self, item):
        return self.drill_path[item]

    def __setitem__(self, key, value):
        self.drill_path[key] = value

    def __str__(self):
        return str(self.drill_path)

    def __repr__(self):
        return repr(self.drill_path)

    def __iter__(self):
        return iter(self.drill_path)

    def __eq__(self, other):
        if isinstance(other, DrillPaths):
            other = other.drill_path
        for item in self.drill_path:
            if item not in other:
                return False
        for item in other:
            if item not in self.drill_path:
                return False
        return True

    def get(self, drill_path):
        """ Get the Drill Path object from DrillPaths

        Args:
            drill_path (str): The drill path to get

        Returns: The Drill Path object
        """
        return self.drill_path.get(drill_path)

    def add(self, drill_path):
        """ Add the Drill Path object to DrillPaths

        Args:
            drill_path (str): The drill path to add
        """
        self.drill_path.add(drill_path)

    def delete(self, drill_path):
        """ Delete the Drill Path object from DrillPaths

        Args:
            drill_path (str): The drill path to delete
        """
        self.drill_path.delete(drill_path)

    def update(self, drill_path):
        """ Update the DrillPaths object

        Args:
            drill_path (str): The drill path to update
        """
        self.drill_path.update(drill_path)

    def dict(self):
        return {'drill-path': [f.dict() for f in self.drill_path]}


@dataclass
class Connection(TableauFileObject):
    """ The Connection Tableau file object """
    tag: str = 'connection'
    authentication: str = None
    class_name: str = None
    dbname: str = None
    schema: str = None
    server: str = None
    service: str = None
    username: str = None
    warehouse: str = None
    instanceurl: str = None
    odbc_connect_string_extras: str = None
    one_time_sql: str = None
    role: str = None
    query_tagging_enabled: bool = None
    saml_idp: str = None
    server_oauth: str = None
    server_userid: str = None
    workgroup_auth_mode: str = None
    tablename: str = None
    default_settings: str = None
    directory: str = None
    filename: str = None
    extract_engine: bool = None
    port: int = None

    def dict(self):
        output = dict()
        if self.authentication is not None:
            output['@authentication'] = self.authentication
        if self.class_name is not None:
            output['@class'] = self.class_name
        if self.dbname is not None:
            output['@dbname'] = self.dbname
        if self.schema is not None:
            output['@schema'] = self.schema
        if self.server is not None:
            output['@server'] = self.server
        if self.service is not None:
            output['@service'] = self.service
        if self.username is not None:
            output['@username'] = self.username
        if self.warehouse is not None:
            output['@warehouse'] = self.warehouse
        if self.instanceurl is not None:
            output['@instanceurl'] = self.instanceurl
        if self.odbc_connect_string_extras is not None:
            output['@odbc-connect-string-extras'] = self.odbc_connect_string_extras
        if self.one_time_sql is not None:
            output['@one-time-sql'] = self.one_time_sql
        if self.role is not None:
            output['@role'] = self.role
        if self.query_tagging_enabled is not None:
            output['@query-tagging-enabled'] = self.query_tagging_enabled
        if self.saml_idp is not None:
            output['@saml-idp'] = self.saml_idp
        if self.server_oauth is not None:
            output['@server-oauth'] = self.server_oauth
        if self.server_userid is not None:
            output['@server-userid'] = self.server_userid
        if self.workgroup_auth_mode is not None:
            output['@workgroup-auth-mode'] = self.workgroup_auth_mode
        if self.tablename is not None:
            output['@tablename'] = self.tablename
        if self.default_settings:
            output['@default-settings'] = self.default_settings
        if self.directory is not None:
            output['@directory'] = self.directory
        if self.filename is not None:
            output['@filename'] = self.filename
        if self.extract_engine is not None:
            output['@extract-engine'] = str(self.extract_engine).lower()
        if self.port is not None:
            output['@port'] = str(self.port)
        return output


@dataclass
class NamedConnection(TableauFileObject):
    """ The NamedConnection Tableau file object """
    name: str
    tag: str = 'named-connection'
    caption: str = None
    connection: Connection = None

    def __post_init__(self):
        if self.connection:
            self.connection = Connection(**transform_tableau_object(self.connection))
        super().__post_init__()

    def __hash__(self):
        return hash(str(astuple(self)))

    def __eq__(self, other):
        if isinstance(other, str):
            return self.connection.class_name == other
        if isinstance(other, dict):
            return self.connection.class_name == other.get('connection', {}).get('class_name')
        if isinstance(other, (NamedConnection, object)) and other.connection is not None:
            if self.connection is None:
                return other.connection is None
            return other.connection is not None and self.connection.class_name == other.connection.class_name
        return False

    def dict(self):
        return {
            '@caption': self.caption,
            '@name': self.name,
            'connection': self.connection.dict()
        }


@dataclass
class MetadataRecord(TableauFileObject):
    """ The MetadataRecord Tableau file object """
    class_name: str
    remote_name: str
    remote_type: str
    parent_name: str
    remote_alias: str
    tag: str = 'metadata-record'
    ordinal: int = None
    family: str = None
    local_type: str = None
    local_name: str = None
    aggregation: str = None
    approx_count: int = None
    width: int = None
    precision: int = None
    scale: int = None
    contains_null: bool = None
    collation: dict = None
    attributes: list = None
    object_id: str = None

    def __post_init__(self):
        if self.attributes and isinstance(self.attributes, dict):
            self.attributes = self.attributes['attribute']
        super().__post_init__()

    def __hash__(self):
        return hash(str(astuple(self)))

    def __eq__(self, other):
        if isinstance(other, str):
            return self.remote_name == other
        if isinstance(other, dict):
            return self.remote_name == other.get('remote_name')
        if isinstance(other, (Column, object)):
            return self.remote_name == other.remote_name
        return False

    def dict(self):
        output = {
            '@class': self.class_name,
            'remote-name': self.remote_name,
            'remote-type': self.remote_type,
            'local-name': self.local_name,
            'parent-name': self.parent_name,
            'remote-alias': self.remote_alias,
            'local-type': self.local_type
        }

        if self.object_id is not None:
            output['object-id'] = self.object_id
        if self.ordinal is not None:
            output['ordinal'] = self.ordinal
        if self.aggregation is not None:
            output['aggregation'] = self.aggregation
        if self.family is not None:
            output['family'] = self.family
        if self.contains_null is not None:
            output['contains-null'] = str(self.contains_null).lower()
        if self.approx_count is not None:
            output['approx-count'] = self.approx_count
        if self.scale is not None:
            output['scale'] = self.scale
        if self.precision is not None:
            output['precision'] = self.precision
        if self.width is not None:
            output['width'] = self.width
        if self.collation is not None:
            output['collation'] = self.collation
        if self.attributes is not None:
            output['attributes'] = dict()
            output['attributes']['attribute'] = self.attributes
        return output


@dataclass
class RefreshEvent(TableauFileObject):
    """ The RefreshEvent Tableau file object """
    add_from_file_path: str = None
    increment_value: str = None
    refresh_type: str = None
    rows_inserted: str = None
    timestamp_start: datetime = None

    def dict(self):
        dictionary = dict()
        if self.add_from_file_path is not None:
            dictionary['@add-from-file-path'] = self.add_from_file_path
        if self.increment_value is not None:
            dictionary['@increment-value'] = self.increment_value
        if self.refresh_type is not None:
            dictionary['@refresh-type'] = self.refresh_type
        if self.rows_inserted is not None:
            dictionary['@rows-inserted'] = self.rows_inserted
        if isinstance(self.timestamp_start, datetime):
            dictionary['@timestamp-start'] = self.timestamp_start.strftime('%Y-%m-%d %H:%M:%S.%f')
        return dictionary


@dataclass
class Refresh(TableauFileObject):
    """ The Refresh Tableau file object """
    tag: str = 'refresh'
    refresh_event: RefreshEvent = None
    increment_key: str = None
    incremental_updates: bool = None

    def __post_init__(self):
        if isinstance(self.refresh_event, dict):
            self.refresh_event = RefreshEvent(**transform_tableau_object(self.refresh_event))
        super().__post_init__()

    def dict(self):
        dictionary = dict()
        if isinstance(self.increment_key, str):
            dictionary['@increment-key'] = self.increment_key
        if isinstance(self.incremental_updates, bool):
            dictionary['@incremental-updates'] = str(self.incremental_updates).lower()
        if isinstance(self.refresh_event, RefreshEvent):
            dictionary['refresh-event'] = self.refresh_event.dict()
        else:
            dictionary['@refresh-event'] = self.refresh_event
        return dictionary


@dataclass
class ParentConnection(TableauFileObject):
    """ The parent Connection Tableau file object """
    tag: str = 'connection'
    class_name: str = None
    authentication: str = None
    access_mode: str = None
    author_locale: str = None
    dbname: str = None
    schema: str = None
    sslmode: str = None
    tablename: str = None
    username: str = None
    update_time: str = None
    extract_engine: bool = None
    named_connections: TableauFileObjects[NamedConnection] = None
    relation: Relation = None
    cols: TableauFileObjects[MappingCol] = None
    refresh: Refresh = None
    metadata_records: TableauFileObjects[MetadataRecord] = None
    default_settings: str = None

    def __post_init__(self):
        if self.refresh is not None:
            self.refresh = Refresh(**transform_tableau_object(self.refresh))
        if self.relation is not None:
            self.relation = Relation(**transform_tableau_object(self.relation))
        if self.named_connections is not None:
            self.named_connections = TableauFileObjects(
                self.named_connections['named-connection'], item_class=NamedConnection, tag='named-connections')
        else:
            self.named_connections = TableauFileObjects(item_class=NamedConnection, tag='named-connections')
        if self.cols is not None:
            self.cols = TableauFileObjects(self.cols['map'], item_class=MappingCol, tag='cols')
        else:
            self.cols = TableauFileObjects(item_class=MappingCol, tag='cols')
        if self.metadata_records is not None:
            self.metadata_records = TableauFileObjects(
                self.metadata_records['metadata-record'], item_class=MetadataRecord, tag='metadata-records'
            )
        else:
            self.metadata_records = TableauFileObjects(item_class=MetadataRecord, tag='metadata-records')
        super().__post_init__()

    def __getitem__(self, item):
        if item in self.named_connections:
            return self.named_connections[item].connection

    def __setitem__(self, key, value):
        self.named_connections[key].connection = value

    def get(self, item):
        """ Gets the TableauFileObject Connection object from the named_connection

        Args:
            item (NamedConnection|Connection|str|int): A TableauFileObject Connection object; class_name of the object;
             or index of the object

        Returns: The TableauFileObject Connection object
        """
        if isinstance(item, Connection) and item.class_name in self.named_connections:
            return self.named_connections[item.class_name].connection
        if isinstance(item, (NamedConnection, str, int)) and item in self.named_connections:
            return self.named_connections[item].connection

    def update(self, item):
        """ Updates the TableauFileObject Connection object of the named_connection

        Args:
            item (Connection): A TableauFileObject Connection object

        Returns: The TableauFileObject Connection object
        """
        if item.class_name in self.named_connections:
            self.named_connections[item.class_name].caption = item.server
            self.named_connections[item.class_name].connection = item
            if item.schema and self.relation.type == 'table':
                self.relation.table = f'[{item.schema}].' + self.relation.table.split('.')[-1]

    def dict(self):
        dictionary = {'@class': self.class_name}
        if self.named_connections:
            dictionary['named-connections'] = {'named-connection': [nc.dict() for nc in self.named_connections]}
        if self.relation is not None:
            dictionary['relation'] = self.relation.dict()
        if self.cols is not None:
            dictionary['cols'] = {'map': [c.dict() for c in self.cols]}
        if self.refresh is not None:
            dictionary['refresh'] = self.refresh.dict()
        if self.metadata_records is not None:
            dictionary['metadata-records'] = {'metadata-record': [m.dict() for m in self.metadata_records]}
        if self.authentication is not None:
            dictionary['@authentication'] = self.authentication
        if self.access_mode is not None:
            dictionary['@access-mode'] = self.access_mode
        if self.author_locale is not None:
            dictionary['@author-locale'] = self.author_locale
        if self.dbname is not None:
            dictionary['@dbname'] = self.dbname
        if self.schema is not None:
            dictionary['@schema'] = self.schema
        if self.sslmode is not None:
            dictionary['@sslmode'] = self.sslmode
        if self.tablename is not None:
            dictionary['@tablename'] = self.tablename
        if self.username is not None:
            dictionary['@username'] = self.username
        if self.update_time is not None:
            dictionary['@update-time'] = self.update_time
        if self.extract_engine is not None:
            dictionary['@extract-engine'] = str(self.extract_engine).lower()
        return dictionary


@dataclass
class Extract(TableauFileObject):
    """ The Extract Tableau file object """
    object_id: str = None
    user_specific: bool = None
    count: int = None
    enabled: bool = None
    units: str = None
    connection: ParentConnection = None
    tag: str = 'extract'

    def __post_init__(self):
        if self.connection is not None:
            self.connection = ParentConnection(**transform_tableau_object(self.connection))
        super().__post_init__()

    def dict(self):
        dictionary = dict()
        if self.object_id is not None:
            dictionary['@object-id'] = self.object_id
        if self.user_specific is not None:
            dictionary['@user-specific'] = str(self.user_specific).lower()
        if self.count is not None:
            dictionary['@count'] = self.count
        if self.enabled is not None:
            dictionary['@enabled'] = str(self.enabled).lower()
        if self.units is not None:
            dictionary['@units'] = self.units
        if self.connection is not None:
            dictionary['connection'] = self.connection.dict()
        return dictionary


@dataclass
class Aliases(TableauFileObject):
    """ The Aliases Tableau file object """
    enabled: bool = True
    tag: str = 'aliases'

    def dict(self):
        return {'@enabled': 'yes' if self.enabled else 'no'}


@dataclass
class DateOptions(TableauFileObject):
    """ The DateOptions Tableau file object """
    fiscal_year_start: str = None
    start_of_week: str = None
    tag: str = 'date-options'

    def dict(self):
        dictionary = dict()
        if self.fiscal_year_start is not None:
            dictionary['@fiscal-year-start'] = self.fiscal_year_start
        if self.fiscal_year_start is not None:
            dictionary['@start-of-week'] = self.start_of_week
        return dictionary


@dataclass
class ColumnInstance(TableauFileObject):
    """ The ColumnInstance Tableau file object """
    column: str = None
    derivation: str = None
    name: str = None
    pivot: str = None
    type: str = None
    tag: str = 'column-instance'

    def dict(self):
        dictionary = dict()
        if self.column is not None:
            dictionary['@column'] = self.column
        if self.derivation is not None:
            dictionary['@derivation'] = self.derivation
        if self.name is not None:
            dictionary['@name'] = self.name
        if self.pivot is not None:
            dictionary['@pivot'] = self.pivot
        if self.type is not None:
            dictionary['@type'] = self.type
        return dictionary


if __name__ == '__main__':
    t1 = Column(name='Test', datatype='integer', role='measure', type='quantitative', calculation='COUNT(1)')
    t2 = Column(name='Test', datatype='string', role='dimension', type='ordinal')
    print(t1 == 'Test')
    print(t1 == {'name': 'Test'})
    print(t1 == t2)
    print(t1.dict())
    print(t2.dict())



import logging
import tableau_utilities.tableau_server.tableau_server_objects as tso
from tableau_utilities.tableau_server.static import TableauConnectionError, transform_tableau_object
from tableau_utilities.tableau_server.base import Base


class Get(Base):
    """ Core Get functionality of the TableauServer class """
    def __init__(self, parent):
        super().__init__(parent)

    def __get_objects_pager(self, url, obj, page_size=100):
        """ GET all objects in the site
        Args:
            url: The url of for the request, i.e /api/api-version/sites/site-id/groups
            obj: The name of the object being requested
            page_size: The size of the page (number of objects per page)
        Returns: A list of the objects
        """
        response = self._get(url)
        total_available = response.get('pagination', {}).get('totalAvailable', page_size)
        total_available = int(total_available)
        current = 0
        page = 0
        while current < total_available:
            current += page_size
            page += 1
            if '?' in url:
                page_url = f'{url}&pageSize={page_size}&pageNumber={page}'
            else:
                page_url = f'{url}?pageSize={page_size}&pageNumber={page}'
            logging.info('GET %s --> %s/%s', page_url, current, total_available)
            res = self._get(page_url)
            for obj_dict in res[f'{obj}s'][obj]:
                transform_tableau_object(obj_dict)
                yield obj_dict

    def datasources(self):
        """ Queries for all datasources in the site
            URI GET /api/api-version/sites/site-id/datasources
        Returns: All datasources in the site
        """
        url = f"{self.url}/datasources?fields=_default_" \
              f",favoritesTotal" \
              f",databaseName" \
              f",connectedWorkbooksCount" \
              f",hasAlert" \
              f",hasExtracts" \
              f",isPublished" \
              f",serverName"

        for d in self.__get_objects_pager(url, 'datasource'):
            yield tso.Datasource(**d)

    def datasource(self, datasource_id=None, datasource_name=None, datasource_project=None):
        """ Queries for a datasource in the site
            URI GET /api/api-version/sites/site-id/datasources/datasource_id

            (Optional) Can get the datasource either by ID, or by name & project.

        Args:
              datasource_id (str): The ID of the datasource
              datasource_name (str): The name of the datasource
              datasource_project (str): The name of the project the datasource is in

        Returns: A Datasource Tableau object
        """
        if datasource_id:
            d = self._get(f'{self.url}/datasources/{datasource_id}')
            d = d['datasource']
            transform_tableau_object(d)
            return tso.Datasource(**d)
        elif datasource_name and datasource_project:
            for d in self.datasources():
                if d.name == datasource_name and d.project_name == datasource_project:
                    return d
            raise TableauConnectionError(
                f'Datasource not found:\n\tName    -> {datasource_name}\n\tProject -> {datasource_project}'
            )

        raise TableauConnectionError(
            'Please provide either the datasource_id, or both datasource_name and datasource_project'
        )

    def datasource_connections(self, datasource_id):
        """ Queries for all Connection Tableau objects in the datasource
            URI GET /api/api-version/sites/site-id/datasources/datasource_id/connections
        Args:
            datasource_id (str): The ID of the Tableau Datasource
        Returns: All Connection Tableau objects in the datasource
        """
        url = f'{self.url}/datasources/{datasource_id}/connections'
        for connection in self.__get_objects_pager(url, 'connection'):
            yield tso.Connection(**connection)

    def workbook_connections(self, workbook_id):
        """ Queries for all Connection Tableau objects in the workbook
            URI GET /api/api-version/sites/site-id/workbooks/workbook_id/connections
        Args:
            workbook_id (str): The ID of the Tableau Workbook
        Returns: All Connection Tableau objects in the workbook
        """
        url = f'{self.url}/workbooks/{workbook_id}/connections'
        for connection in self.__get_objects_pager(url, 'connection'):
            yield tso.Connection(**connection)

    def workbooks(self):
        """ Queries for all workbooks in the site
            URI GET /api/api-version/sites/site-id/workbooks
        Returns: All workbooks in the site
        """
        for w in self.__get_objects_pager(f"{self.url}/workbooks", 'workbook'):
            yield tso.Workbook(**w)

    def workbook(self, workbook_id=None, workbook_name=None, workbook_project=None):
        """ Queries for a workbook in the site
            URI GET /api/api-version/sites/site-id/workbooks/workbook_id
        Args:
            workbook_id (str): The ID of the Tableau Workbook
            workbook_name (str): The name of the workbook
            workbook_project (str): The name of the project the workbook is in
        Returns: A Workbooks Tableau object
        """
        if workbook_id:
            w = self._get(f'{self.url}/workbooks/{workbook_id}')
            w = w['workbook']
            transform_tableau_object(w)
            return tso.Workbook(**w)
        elif workbook_name and workbook_project:
            for w in self.workbooks():
                if w.name == workbook_name and w.project_name == workbook_project:
                    return w
            raise TableauConnectionError(
                f'Datasource not found:\n\tName    -> {workbook_name}\n\tProject -> {workbook_project}'
            )

        raise TableauConnectionError(
            'Please provide either the workbook_id, or both workbook_name and workbook_project'
        )

    def views(self):
        """ Queries for all views in the site
            URI GET /api/api-version/sites/site-id/views
        Returns: All workbooks in the site
        """
        url = f"{self.url}/views?fields=_default_,sheetType,usage"
        for v in self.__get_objects_pager(url, 'view'):
            yield tso.View(**v)

    def view(self, view_id):
        """ Queries for a view in the site
            URI GET /api/api-version/sites/site-id/views/view_id
        Args:
            view_id (str): The ID of the Tableau View
        Returns: A View Tableau object
        """
        v = self._get(f'{self.url}/views/{view_id}')
        v = v['view']
        return tso.View(**v)

    def projects(self, top_level_only=True):
        """ Queries for all projects in the site
            URI GET /api/api-version/sites/site-id/projects
        Args:
            top_level_only (bool): True to only get top level projects
        Returns: All top level projects in the site
        """
        url = f"{self.url}/projects?fields=_default_" \
              f",topLevelProject" \
              f",writeable" \
              f",contentsCounts.projectCount" \
              f",contentsCounts.viewCount" \
              f",contentsCounts.datasourceCount" \
              f",contentsCounts.workbookCount"
        for p in self.__get_objects_pager(url, 'project'):
            project = tso.Project(**p)
            # Only get Top Level projects
            if top_level_only and not project.top_level_project:
                continue
            yield project

    def project(self, project_id):
        """ Queries for the project by project_id
            URI GET /api/api-version/sites/site-id/projects/project_id
        Args:
            project_id (str): The ID of the project in Tableau Online
        Returns: A Tableau Project object specified by ID
        """
        for p in self.projects():
            if p.id == project_id:
                return p

    def groups(self):
        """ Queries for all groups in the site
            URI GET /api/api-version/sites/site-id/groups
        Returns: All groups in the site
        """
        url = f"{self.url}/groups?fields=_default_,userCount,minimumSiteRole"
        for g in self.__get_objects_pager(url, 'group'):
            yield tso.Group(**g)

    def group(self, group_id):
        """ Queries for the group by user_id
            URI GET /api/api-version/sites/site-id/groups/group_id
        Args:
            group_id (str): The ID of the group in Tableau Online
        Returns: A Tableau Group object specified by ID
        """
        for group in self.groups():
            if group.id == group_id:
                return group

    def users(self):
        """ Queries for all users in the site
            URI GET /api/api-version/sites/site-id/users
        Returns: All users in the site
        """
        url = f"{self.url}/users?fields=_default_,fullName,email"
        for user in self.__get_objects_pager(url, 'user'):
            yield tso.User(**user)

    def user(self, user_id):
        """ Queries for the user by user_id
            URI GET /api/api-version/sites/site-id/users/user_id
        Args:
            user_id (str): The ID of the user in Tableau Online
        Returns: A Tableau User object specified by ID
        """
        u = self._get(f"{self.url}/users/{user_id}")
        u = u['user']
        transform_tableau_object(u)
        return tso.User(**u)

    def user_groups(self):
        """ Queries for all groups and all user in those groups
            URI GET /api/api-version/sites/site-id/groups/group_id/users
        Returns: A list of all user/group combinations
        """
        for group in self.groups():
            for u in self.__get_objects_pager(f"{self.url}/groups/{group.id}/users", 'user'):
                yield group, tso.User(**u)



import cgi
import os
import requests
from requests import Session
from tableau_utilities.tableau_server.static import TableauConnectionError
from tableau_utilities.tableau_server.base import Base


class Download(Base):
    """ Core Download functionality of the TableauServer class """
    def __init__(self, parent):
        super().__init__(parent)

    def __download_object(self, url, file_dir=None):
        """ Downloads a datasource from Tableau Online

        Args:
            url (str): The URL for the request
            file_dir (str): The file directory to write the file to

        Returns: The absolute path to the file
        """
        res = self.session.get(url, stream=True)
        try:
            res.raise_for_status()
        except requests.exceptions.HTTPError as err:
            raise TableauConnectionError(err) from err
        _, params = cgi.parse_header(res.headers["Content-Disposition"])
        file_name = os.path.basename(params["filename"])
        if file_dir:
            os.makedirs(file_dir, exist_ok=True)
            path = os.path.join(file_dir, file_name)
        else:
            path = file_name
        with open(path, "wb") as f:
            # Download in 1024 bytes (1kb) chunks
            for chunk in res.iter_content(1024):
                f.write(chunk)
        return os.path.abspath(path)

    def datasource(self, datasource_id, file_dir=None, include_extract=False):
        """ Downloads a datasource from Tableau Online

        Args:
            datasource_id (str):
            file_dir (str):
            include_extract (bool):
        """
        return self.__download_object(
            f'{self.url}/datasources/{datasource_id}/content?includeExtract={include_extract}',
            file_dir
        )

    def workbook(self, workbook_id, file_dir=None, include_extract=False):
        """ Downloads a workbook from Tableau Online

        Args:
            workbook_id (str):
            file_dir (str):
            include_extract (bool):
        """
        return self.__download_object(
            f'{self.url}/workbooks/{workbook_id}/content?includeExtract={include_extract}',
            file_dir
        )



import json
from pathlib import Path

import pandas as pd

from tableau_utilities.general.cli_styling import Color, Symbol
from tableau_utilities.general.config_column_persona import get_persona_by_attribs, get_persona_by_metadata_local_type
from tableau_utilities.general.funcs import convert_to_snake_case
from tableau_utilities.tableau_file.tableau_file import Datasource
from tableau_utilities.tableau_server.tableau_server import TableauServer


def load_csv_with_definitions(file=None):
    """ Returns a dictionary with the definitions from a csv. The columns are expected to include column_name and description

    Args:
        file: The path to the .csv file with the definitions. The csv must include a column_name and description.

    Returns:
        dictionary mapping column name to definition

    """

    definitions_mapping = dict()
    df = pd.read_csv(file)
    df.columns = df.columns.str.lower()
    definitions = df.to_dict('records')

    # Check that the csv contains column_name and description headers
    column_names = list(df.columns)
    if 'column_name' not in column_names or 'description' not in column_names:
        raise ValueError('The .csv must contain a column_name and a description column.')

    for column in definitions:
        if str(column['description']) != 'nan':
            definitions_mapping[column['column_name']] = column['description']

    return definitions_mapping


def choose_persona(role, role_type, datatype, caption):
    """  The config relies on a persona which is a combination of role, role_type and datatype for each column.
    This returns the persona name or raises an exception if the combination is not found

    Args:
        role: dimension or measure
        role_type: nominal, ordinal, or quantitative
        datatype: string, date, datetype, real, or boolean
        caption: The name of the column for the persona. Used in error message

    """
    p = get_persona_by_attribs(role, role_type, datatype)
    if p:
        return p

    raise ValueError(
        f"There is no persona for the combination of ROLE {role}, ROLE_TYPE {role_type}, and DATATYPE {datatype}."
        f"Error on column {caption}'"
    )


def get_metadata_record_config(metadata_records, datasource_name, debugging_logs=False):
    """ Builds a column config for columns that are only in metadata records and not in column objects

    Args:
        metadata_records (Datasource.connection.metadata_records): The metadata_records list from the Datasource object
        datasource_name (str): The name of the datasource
        debugging_logs: Prints information to consolde if true
    """
    metadata_record_columns = dict()

    for m in metadata_records:
        if m.class_name == 'column':
            local_name = m.local_name[1:-1]
            persona = get_persona_by_metadata_local_type(m.local_type) or ''

            metadata_record_columns[local_name] = {
                'persona': persona,
                "datasources": [
                    {
                        "name": datasource_name,
                        "local-name": local_name,
                        "sql_alias": m.remote_name
                    },
                ]
            }

            if debugging_logs:
                print(m)
                print(persona)
                print(local_name, ':', m.remote_name)
                print(metadata_record_columns[local_name])

    return metadata_record_columns


def create_column_config(columns, datasource_name, folder_mapping, metadata_record_columns, definitions_mapping,
                         debugging_logs):
    """ Generates a list of column configs with None for a folder

    Args:
        columns (Datasource.columns): The list of Column objects from the Datasource object
        datasource_name (str): The name of the datasource
        folder_mapping (dict): A dict mapping column name to folder name
        metadata_record_columns (dict): The config based on metadata records; used if there is no column object
        definitions_mapping (dict): The mapping of definitions from a csv
        debugging_logs (bool): True to print debugging logs

    Returns:
        column_config (dict): A dict config for fields that are pulled from the source database
        calculated_column_configs (dict): A dictionary with the configs for calculated fields added in the workbook that built
            the datasource

        Sample column_config Output: {
          "Salesforce Opportunity Id": {
            "description": "The 18 digit account Id for a Salesforce opportunity",
            "folder": Name,
            "persona": "string_dimension",
            "datasources": [
              {
                "name": "Opportunity",
                "local-name": "SALESFORCE_OPPORTUNITY_ID",
                "sql_alias": "SALESFORCE_OPPORTUNITY_ID"
              }
            ]
        }
    """

    column_config = dict()
    calculated_column_configs = dict()
    column_name_list = list()

    for column in columns:
        column_name = column.name[1:-1]

        # Skip internal object columns
        if column_name.startswith('__tableau_internal_object_id__'):
            if debugging_logs:
                print(f'Skipping {column_name}: starts with __tableau_internal_object_id__ ')
            continue
        # Skip columns with the 'table' datatype for now
        if column.datatype == 'table':
            if debugging_logs:
                print(f'Skipping {column_name}: column.datatype == "table" ')
            continue

        # Keeps a list of column names from the column object.
        column_name_list.append(column_name)

        # Make a title case caption from the database name if there is no caption
        if column.caption:
            caption = column.caption
        else:
            caption = column_name.replace('_', ' ').title()

        persona = choose_persona(role=column.role, role_type=column.type, datatype=column.datatype,
                                 caption=caption)

        # Takes the description from the csv if there is one
        # Assumes the csv  is the source of truth if there are definitions in both
        if definitions_mapping.get(caption, '') != '':
            description = definitions_mapping[caption]
        else:
            description = column.desc or ''

        folder_name = folder_mapping.get(column_name)

        column_dict = {
            "description": description,
            "folder": folder_name,
            "persona": persona,
            "datasources": [
                {
                    "name": datasource_name,
                    "local-name": column_name
                },
            ]
        }
        # Optional Properties to Add
        if column.fiscal_year_start:
            column_dict['fiscal_year_start'] = column.fiscal_year_start
        if column.default_format:
            column_dict['default_format'] = column.default_format

        if debugging_logs:
            print('-' * 30)
            print(caption)
            print(column.dict())
        # Calculations are written to a separate config in the Airflow DAG
        if column.calculation:
            calculated_column_configs[caption] = column_dict
            calculated_column_configs[caption]["calculation"] = column.calculation
            if debugging_logs:
                print('CALCULATED COLUMN CONFIG FIELD')
                print(calculated_column_configs[caption])
        else:
            if column_name in metadata_record_columns:
                sql_alias = metadata_record_columns[column_name]['datasources'][0]['sql_alias']
            else:
                sql_alias = column_name
            column_config[caption] = column_dict
            column_config[caption]['datasources'][0]['sql_alias'] = sql_alias
            if debugging_logs:
                print('COLUMN CONFIG FIELD')
                print(column_config[caption])

    # Add column configs for metadata_record columns when there wasn't a column object already
    # This is only need for non-calulated fields
    for k, v in metadata_record_columns.items():
        if k in column_name_list:
            continue

        caption = v['datasources'][0]['sql_alias'].replace('_', ' ').title()

        if v['persona']:
            column_config[caption] = {
                "description": '',
                "folder": None,
                "persona": v['persona'],
                "datasources": v['datasources']
            }

            if debugging_logs:
                print('-' * 30)
                print('METADATA RECORD COLUMN ADDED')
                print(caption)
                print(v['persona'])
                print(column_config[caption])
        else:
            if debugging_logs:
                print('-' * 30)
                print('METADATA RECORD COLUMN SKIPPED - Missing persona')
                print(caption)
                print(k, ':', v)

    return column_config, calculated_column_configs


def build_folder_mapping(folders):
    """  Builds a dictionary mapping columns to folders

    Args:
        folders (Datasource.folders_common): List of folder objects from the Datasource object

    Returns: A dict of column & folder mapping; i.e. {'column1': 'folderA', 'column2': 'folderA'}
    """
    mappings = dict()
    for folder in folders:
        if folder.folder_item:
            for item in folder.folder_item:
                field_name = item.name[1:-1]
                mappings[field_name] = folder.name
    return mappings


def generate_config(args, server: TableauServer = None):
    """ Downloads a datasource and saves configs for that datasource

    """

    # Set variables from the args
    definitions_csv_path = args.definitions_csv
    debugging_logs = args.debugging_logs
    location = args.location
    id = args.id
    datasource_name = args.name
    datasource_path = args.file_path
    project_name = args.project_name

    # Print Styling
    color = Color()
    symbol = Symbol()

    # Set file_prefix to false when this function is called from the merge_config namespace
    try:
        file_prefix = args.file_prefix
    except:
        file_prefix = False

    # Get the datasource name from the path
    if location == 'local':
        datasource_name = Path(datasource_path).stem
    # Download the datasouce and set values for
    elif location == 'online':
        obj = server.get.datasource(id, datasource_name, project_name)
        id = obj.id
        datasource_name = obj.name
        print(f'{color.fg_yellow}GETTING DATASOURCE {symbol.arrow_r} '
              f'{color.fg_grey}ID: {id} {symbol.sep} '
              f'NAME: {datasource_name} {symbol.sep} '
              f'INCLUDE EXTRACT: false{color.reset}')
        datasource_path = server.download.datasource(id, include_extract=False)

    print(f'{color.fg_yellow}BUILDING CONFIG {symbol.arrow_r} '
          f'{color.fg_grey}{datasource_name} {symbol.sep} {datasource_path}{color.reset}')
    datasource = Datasource(datasource_path)
    # Get column information from the metadata records
    metadata_record_config = get_metadata_record_config(
        datasource.connection.metadata_records,
        datasource_name,
        debugging_logs
    )

    # Get the mapping of definitions from the csv
    definitions_mapping = dict()
    if definitions_csv_path is not None:
        definitions_mapping = load_csv_with_definitions(file=definitions_csv_path)

    # Extract the columns and folders. Build the new config
    folder_mapping = build_folder_mapping(datasource.folders_common)
    column_configs, calculated_column_configs = create_column_config(
        columns=datasource.columns,
        datasource_name=datasource_name,
        folder_mapping=folder_mapping,
        metadata_record_columns=metadata_record_config,
        definitions_mapping=definitions_mapping,
        debugging_logs=debugging_logs
    )

    # Sort configs
    column_configs = dict(sorted(column_configs.items()))
    calculated_column_configs = dict(sorted(calculated_column_configs.items()))

    datasource_name_snake = convert_to_snake_case(datasource_name)
    output_file_column_config = 'column_config.json'
    output_file_calculated_column_config = 'tableau_calc_config.json'

    if file_prefix:
        output_file_column_config = f'{datasource_name_snake}__{output_file_column_config}'
        output_file_calculated_column_config = f'{datasource_name_snake}__{output_file_calculated_column_config}'

    with open(output_file_column_config, "w") as outfile:
        json.dump(column_configs, outfile)

    with open(output_file_calculated_column_config, "w") as outfile:
        json.dump(calculated_column_configs, outfile)

    print(f'{color.fg_green}{symbol.success}  COLUMN CONFIG {symbol.arrow_r} '
          f'{color.fg_grey}{output_file_column_config}{color.reset}')
    print(f'{color.fg_green}{symbol.success}  CALCULATED COLUMN CONFIG {symbol.arrow_r} '
          f'{color.fg_grey}{output_file_calculated_column_config}{color.reset}')

    return output_file_column_config, output_file_calculated_column_config



from tableau_utilities.scripts.merge_config import read_file
import pandas as pd


def csv_config(args):

    # Set variables from args
    debugging_logs = args.debugging_logs
    config_list = args.config_list

    all_columns = []

    for each_config in config_list:
        config = read_file(each_config)

        # columns = []
        for column, details in config.items():
            if debugging_logs:
                print(column, details)

            column_name = column
            description = details['description']
            folder = details['folder']
            persona = details['persona']

            calculation = details.get('calculation')
            fiscal_year_start = details.get('fiscal_year_start')
            default_format = details.get('default_format')

            for datasource in details['datasources']:
                datasource_name = datasource['name']
                local_name = datasource['local-name']
                sql_alias = datasource.get('sql_alias')

                each_column = {'column_name': column_name,
                               'description': description,
                               'folder': folder,
                               'persona': persona,
                               'calculation': calculation,
                               'fiscal_year_start': fiscal_year_start,
                               'default_format': default_format,
                               'datasource_name': datasource_name,
                               'local_name': local_name,
                               'sql_alias': sql_alias
                               }

                all_columns.append(each_column)

    df = pd.DataFrame(all_columns)
    df.to_csv('config_columns.csv', index=False)





import json
import argparse
import os
import yaml
import shutil
from tableau_utilities import Datasource, TableauServer


def do_args():
    """ Parse arguments.

    Returns: an argparse.Namespace
    """

    parser = argparse.ArgumentParser(description='List columns in Tableau datasources')
    parser.add_argument(
        '--server',
        help='Tableau Server URL. i.e. <server_address> in https://<server_address>.online.tableau.com',
        default=None
    )
    parser.add_argument(
        '--site',
        help='Site name. i.e. <site> in https://<server_address>.online.tableau.com/#/site/<site>',
        default=None)
    parser.add_argument('--api_version', help='Tableau API version', default='2.8')
    parser.add_argument('--user', required=True, help='user name')
    parser.add_argument('--password', required=True, help='password')
    return parser.parse_args()


def all_columns_all_datasources(server):
    """ Gets a list of all columns in all datasources

    Args:
        server (TableauServer): A Tableau server object
    """
    shutil.rmtree('tmp_tdsx', ignore_errors=True)
    tmp_folder = 'tmp_tdsx'
    os.makedirs(tmp_folder, exist_ok=True)
    os.chdir(tmp_folder)
    datasource_list = [d for d in server.get.datasources()]
    rows = dict()
    for datasource in datasource_list:
        print(datasource.project_name, (datasource.id, datasource.name))
        datasource_path = server.download.datasource(datasource.id, include_extract=False)
        columns = [c.dict() for c in Datasource(datasource_path).columns]
        rows.setdefault(datasource.name, [])
        rows[datasource.name].extend(columns)
    os.chdir('..')
    shutil.rmtree(tmp_folder)
    return rows


if __name__ == '__main__':
    args = do_args()
    with open('settings.yaml') as f:
        settings = yaml.safe_load(f)
    if args.server:
        host = f'https://{args.server}.online.tableau.com'
    else:
        host = settings['tableau_login']['host']
    ts = TableauServer(
        user=args.user or settings['tableau_login']['user'],
        password=args.password or settings['tableau_login']['password'],
        site=args.site or settings['tableau_login']['site'],
        host=host,
        api_version=args.api_version or settings['tableau_login']['api_version']
    )
    config = all_columns_all_datasources(ts)
    with open('generated_config.json', 'w') as fd:
        json.dump(config, fd, indent=3)



""" Static functionality of the TableauServer and Core classes """
import requests
from tableau_utilities.general.funcs import flatten_dict


class TableauConnectionError(Exception):
    """ An Exception in the TableauServer connection """
    pass


def bytes_to_mb(b):
    """ Converts Bytes to Megabytes """
    return b / 1024 / 1024


def mb_to_bytes(mb):
    """ Converts Megabytes to Bytes """
    return mb * 1024 * 1024


def transform_tableau_object(object_dict):
    """ Transforms the object dict from a Tableau REST API call
    Args:
        object_dict (dict): The object dict from a Tableau REST API call
    """
    if object_dict.get('tags'):
        object_dict['tags'] = [t['label'] for t in object_dict['tags']['tag']]
    update = dict()
    flatten_dict(object_dict, update)
    object_dict.clear()
    object_dict.update(update)


def validate_response(response):
    """ Validates the response received from an API call

    Args:
        response (requests.Response): A requests Response object
        raise_for_status (bool): True to raise an error on a bad response

    Returns: The response content as a JSON dict
    """
    info = response.json()
    try:
        response.raise_for_status()
    except requests.exceptions.HTTPError as err:
        error = info.get('error', {})
        raise TableauConnectionError(
            f'\nError: {error.get("code")}: {error.get("summary")} - {error.get("detail")}\n{err}'
        ) from err
    return info



import tableau_utilities.tableau_server.tableau_server_objects as tso
from requests import Session
from tableau_utilities.tableau_server.static import transform_tableau_object
from tableau_utilities.tableau_server.base import Base


class Refresh(Base):
    """ Core Refresh functionality of the TableauServer class """
    def __init__(self, parent):
        super().__init__(parent)

    def datasource(self, datasource_id):
        """ Refresh a datasource """
        content = self._post(f'{self.url}/datasources/{datasource_id}/refresh', json={})
        transform_tableau_object(content['job'])
        return tso.Job(**content['job'])

    def workbook(self, workbook_id):
        """ Refresh a workbook """
        content = self._post(f'{self.url}/workbooks/{workbook_id}/refresh', json={})
        transform_tableau_object(content['job'])
        return tso.Job(**content['job'])



"""
Generic Python functions
"""
import re
from copy import deepcopy


def convert_to_snake_case(string):
    """ Converts a string to snake_case
    Args:
        string (str): A string to convert
    Returns: The transformed string
    """
    # Convert '_This...Thing+123 here__' to 'This_Thing_123_here'
    string = re.sub(r'[^a-zA-Z\d]+', ' ', string)
    string = string.strip().replace(' ', '_')
    # Convert 'ThisThing' to 'this_thing'
    string = re.sub(r'([a-z])([A-Z])', r'\1_\2', string).lower()
    return string


def flatten_dict(dictionary, final_dict, prior_key=None):
    """ Flattens a dictionary.
        Updates the keys to a snake_case in the path of keys flattened.
    Sample:
        - Input:
         {'first_level1': {'second_level1': 123, 'second_level2': 456}, 'first_level2': 'value'}
        - Output:
         {'first_level1_second_level1': 123, 'first_level1_second_level2': 456, 'first_level2': 'value'}
    Args:
        dictionary (dict): The dictionary to flatten
        final_dict (dict): A dictionary to add the flattened values to
        prior_key (str): Should not be given when first calling flatten_dict
    Returns: The flattened dictionary
    """
    for k, v in dictionary.items():
        # Combine the prior_key, with the key in the loop
        current_key = f'{prior_key}_{k}' if prior_key else k
        current_key = convert_to_snake_case(current_key)
        if isinstance(v, dict):
            flatten_dict(v, final_dict, current_key)
        else:
            if current_key not in final_dict:
                final_dict[current_key] = v
            else:
                raise KeyError(f'Attempted to flatten: {current_key}, already in {dictionary}')


def transform_tableau_object(obj):
    """ Transform a Tableau item.

    Args:
        obj (dict): The dict Tableau object
    """
    # Transform the item dict
    update = dict()
    for key, value in obj.items():
        # Update item keys "@someThing-like...this" to "this"
        key = re.sub(r'.+\.\.\.', '', key)
        key = key.replace('ns0_', '').replace('_ns0', '')
        key = convert_to_snake_case(key)
        if key == 'class':
            update['class_name'] = value
        else:
            update[key] = deepcopy(value)
    return update



import pandas as pd
from tabulate import tabulate
from pprint import pprint

from tableau_utilities.general.cli_styling import Color, Symbol
from tableau_utilities.tableau_server.tableau_server import TableauServer


def server_info(args, server):
    """ Prints information for datasources, projects, or workbooks

    Args:
        args: The args from the CLI
        server (TableauServer): the Tableau Server authentication object

    """

    # Set variables from args
    format = args.list_format
    sort_field = args.list_sort_field
    list_object = args.list_object

    # Print Styling
    color = Color()
    symbol = Symbol()

    if list_object:
        print(f'{color.fg_cyan}{list_object.title()}s:{color.reset}')
        # Get server objects, and convert them to dict
        object_list = [o.__dict__ for o in getattr(server.get, f'{list_object.lower()}s')()]
        sorted_records = sorted(object_list, key=lambda d: d[sort_field])
        if format == 'names':
            for record in sorted_records:
                print(record['name'])
        elif format == 'names_ids':
            for record in sorted_records:
                print(record['name'], record['id'])
        elif format == 'ids_names':
            for record in sorted_records:
                print(record['id'], record['name'])
        elif format == 'names_projects':
            for record in sorted_records:
                print(record['name'], record['project_name'])
        elif format == 'full_df':
            df = pd.DataFrame(sorted_records)
            print(tabulate(df, headers='keys', tablefmt='psql', colalign='left'))
        elif format == 'full_dictionary':
            for record in sorted_records:
                print(record)
        elif format == 'full_dictionary_pretty':
            for record in sorted_records:
                pprint(record)



import os
import logging
from time import time
from urllib3.fields import RequestField
from urllib3.filepost import encode_multipart_formdata
import tableau_utilities.tableau_server.tableau_server_objects as tso
from requests import Session
from tableau_utilities.tableau_server.static import (
    TableauConnectionError, bytes_to_mb, mb_to_bytes, transform_tableau_object)
from tableau_utilities.tableau_server.base import Base


class Publish(Base):
    """ Core Publish functionality of the TableauServer class """
    def __init__(self, parent):
        super().__init__(parent)

    @staticmethod
    def __get_multipart_details(parts):
        """ Gets the body and content_type for a multipart/mixed request.

        Args:
            parts (list[tuple[str, str, str, str]]): The parts that make up the RequestField
                i.e. [(name, data, file_name, content_type)]

        Returns: Request body and content_type
        """
        part_list = list()
        for name, data, file_name, content_type in parts:
            r = RequestField(name=name, data=data, filename=file_name)
            r.make_multipart(content_type=content_type)
            part_list.append(r)
        post_body, content_type = encode_multipart_formdata(part_list)
        content_type = ''.join(('multipart/mixed',) + content_type.partition(';')[1:])
        return post_body, content_type

    def __get_datasource_for_publication(self, datasource_id, datasource_name, project_name):
        if datasource_id:
            return self.get.datasource(datasource_id)
        elif datasource_name and project_name:
            project = None
            for p in self.get.projects():
                if p.name == project_name:
                    project = p
                    break
            if not project:
                raise TableauConnectionError(f'Project does not exist: {project}')
            return tso.Datasource(
                name=datasource_name,
                project_id=project.id,
                project_name=project.name
            )
        else:
            raise TableauConnectionError('Specify datasource_id or datasource_name and project_name')

    def __get_workbook_for_publication(self, workbook_id, workbook_name, project_name):
        if workbook_id:
            return self.get.workbook(workbook_id)
        elif workbook_name and project_name:
            project = None
            for p in self.get.projects():
                if p.name == project_name:
                    project = p
                    break
            if not project:
                raise TableauConnectionError(f'Project does not exist: {project}')
            return tso.Workbook(
                name=workbook_name,
                project_id=project.id,
                project_name=project.name
            )
        else:
            raise TableauConnectionError('Specify datasource_id or datasource_name and project_name')

    # 323 seconds at 5 mb, 145 seconds at 50mb
    def __upload_in_chunks(self, file_path, chunk_size_mb=5):
        """ Uplaods a file to Tableau, in chunks.
            - PUT /api/api-version/sites/site-id/fileUploads
            - PUT /api/api-version/sites/site-id/fileUploads/upload_session_id

        Args:
            file_path (str): The path to the file
            chunk_size_mb (int): The chunking size of increments to be uploaded

        Returns: An upload_session_id of the uploaded file
        """
        start = time()
        file_name = os.path.basename(file_path)
        # Initialize file upload session
        res = self._post(f'{self.url}/fileUploads')
        upload_session_id = res['fileUpload']["uploadSessionId"]
        # Read file and append data in chunks
        total = round(bytes_to_mb(os.path.getsize(file_path)), 1)
        current = 0
        file = open(file_path, 'rb')
        while True:
            chunk = file.read(mb_to_bytes(chunk_size_mb))
            if not chunk:
                break
            post_body, content_type = self.__get_multipart_details([
                ('request_payload', '', None, 'text/xml'),
                ('tableau_file', chunk, file_name, 'application/octet-stream')
            ])
            # Log progress every so often
            current += chunk_size_mb
            if current == chunk_size_mb or current % 500 == 0 or current >= total:
                logging.info('({} of {} mb) Uploading {}'.format(
                    current if current < total else total, total, file_path))
            self._put(
                f'{self.url}/fileUploads/{upload_session_id}',
                data=post_body, headers={'Content-Type': content_type}
            )
        file.close()
        logging.info('Uploaded {}: {} mb in {} seconds'.format(file_path, total, round(time() - start)))
        return upload_session_id

    def datasource(self, file_path, datasource_id=None, datasource_name=None, project_name=None, **kw):
        """ Publishes a datasource to Tableau Online.
            One of the following MUST be provided:
              - datasource_id: If the datasource already exists
              - datasource_name AND project name: If this is a new datasource

        Args:
            file_path (str): The path to the datasource file (.tds or .tdsx)
            datasource_id (str): The ID of the datasource in Tableau Online
            datasource_name (str): The name of the Datasource
            project_name (str): The name of the Project in Tableau Online

        Keyword Args:
            overwrite (bool): True to overwrite the datasource, if it exists
            as_job (bool): True to kick this off as an async job in Tableau Online
            append (bool): True to append the data to the datasource in Tableau Online
            connection (dict): A dict of connection credentials to embed in the datasource
                i.e. 'username' and 'password'

        Returns: A Datasource Tableau server object
        """
        overwrite = kw.pop('overwrite', True)
        as_job = kw.pop('as_job', False)
        append = kw.pop('append', False)
        connection = kw.pop('connection', None)
        file_name = os.path.basename(file_path)
        extension = file_path.split('.')[-1]
        datasource = self.__get_datasource_for_publication(datasource_id, datasource_name, project_name)
        ds_xml = datasource.publish_xml(connection)
        # Datasource must be less than 64mb to publish all at once
        if bytes_to_mb(os.path.getsize(file_path)) >= 64:
            upload_session_id = self.__upload_in_chunks(file_path)
            publish_url = f'{self.url}/datasources?uploadSessionId={upload_session_id}' \
                          f'&datasourceType={extension}&overwrite={overwrite}&append={append}&asJob={as_job}'
            post_body, content_type = self.__get_multipart_details([
                ('request_payload', ds_xml, None, 'text/xml')
            ])
        else:
            publish_url = f'{self.url}/datasources?datasourceType={extension}' \
                          f'&overwrite={overwrite}&append={append}&asJob={as_job}'
            with open(file_path, 'rb') as f:
                post_body, content_type = self.__get_multipart_details([
                    ('request_payload', ds_xml, None, 'text/xml'),
                    ('tableau_datasource', f.read(), file_name, 'application/octet-stream')
                ])

        # Finally, publish the file uploaded
        start = time()
        logging.info('Publishing uploaded datasource {}'.format(file_path))
        content = self._post(publish_url, data=post_body, headers={'Content-Type': content_type})
        logging.info('Published uploaded datasource {} in {} seconds'.format(file_path, round(time() - start)))
        transform_tableau_object(content['datasource'])
        return tso.Datasource(**content['datasource'])

    def workbook(self, file_path, workbook_id=None, workbook_name=None, project_name=None, **kw):
        """ Publishes a workbook to Tableau Online.
            One of the following MUST be provided:
              - workbook_id: If the workbook already exists
              - workbook_name AND project name: If this is a new workbook

        Args:
            file_path (str): The path to the Workbook file (.twb or .twbx)
            workbook_id (str): The ID of the Workbook in Tableau Online
            workbook_name (str): The name of the Workbook
            project_name (str): The name of the Project in Tableau Online

        Keyword Args:
            overwrite (bool): True to overwrite the datasource, if it exists
            as_job (bool): True to kick this off as an async job in Tableau Online
            skip_connection_check (bool): True for Tableau server to not check if,
                a non-published connection, of a workbook is reachable
            connections (list[dict]): A list of connections
                i.e. [{address, port, username, password}]

        Returns: A Workbook Tableau server object
        """
        overwrite = kw.pop('overwrite', True)
        as_job = kw.pop('as_job', False)
        skip_connection_check = kw.pop('skip_connection_check', False)
        connections = kw.pop('connections', None)
        file_name = os.path.basename(file_path)
        extension = file_path.split('.')[-1]
        workbook = self.__get_workbook_for_publication(workbook_id, workbook_name, project_name)
        wb_xml = workbook.publish_xml(connections)
        # Datasource must be 64mb or less to publish all at once
        if bytes_to_mb(os.path.getsize(file_path)) > 64:
            upload_session_id = self.__upload_in_chunks(file_path)
            publish_url = f'{self.url}/workbooks?uploadSessionId={upload_session_id}' \
                          f'&workbookType={extension}' \
                          f'&skipConnectionCheck={skip_connection_check}' \
                          f'&overwrite={overwrite}&asJob={as_job}'
            post_body, content_type = self.__get_multipart_details([
                ('request_payload', wb_xml, None, 'text/xml')
            ])
        else:
            publish_url = f'{self.url}/workbooks?workbookType={extension}' \
                          f'&overwrite={overwrite}' \
                          f'&skipConnectionCheck={skip_connection_check}' \
                          f'&asJob={as_job}'
            with open(file_path, 'rb') as f:
                post_body, content_type = self.__get_multipart_details([
                    ('request_payload', wb_xml, None, 'text/xml'),
                    ('tableau_workbook', f.read(), file_name, 'application/octet-stream')
                ])

        # Finally, publish the file uploaded
        start = time()
        logging.info('Publishing uploaded workbook {}'.format(file_path))
        content = self._post(publish_url, data=post_body, headers={'Content-Type': content_type})
        logging.info('Published uploaded workbook {} in {} seconds'.format(file_path, round(time() - start)))
        transform_tableau_object(content['workbook'])
        return tso.Workbook(**content['workbook'])



import webbrowser

from tableau_utilities.general.cli_styling import Color, Symbol, color_print
from tableau_utilities.tableau_server.tableau_server import TableauServer
from tableau_utilities.tableau_server.tableau_server_objects import Datasource, Workbook, Job, Connection


def server_operate(args, server):
    """ Allows downloading, publishing, and refreshing workbooks and datasources

    Args:
        args: The args from the cli
        server (TableauServer): the Tableau Server authentication object

    """

    # Set variables from args
    # Required -> One must be provided
    object_type = args.download or args.publish or args.refresh
    if args.embed_connection:
        object_type = 'datasource'
    # Optional -> Might be None
    include_extract = args.include_extract
    object_id = args.id
    object_name = args.name
    project_name = args.project_name
    all_objects = args.all
    download = args.download
    publish = args.publish
    embed_connection = args.embed_connection
    refresh = args.refresh
    file_path = args.file_path
    connection = None
    if object_type == 'datasource' and args.conn_user and args.conn_pw:
        connection = {'username': args.conn_user, 'password': args.conn_pw}

    # Print Styling
    color = Color()
    symbol = Symbol()

    # Download all objects, and return early if all objects have been downloaded
    if all_objects and download:
        object_list = [o for o in getattr(server.get, f'{object_type}s')()]
        for o in object_list:
            print(
                f'{color.fg_yellow}DOWNLOADING {object_type} {symbol.arrow_r} {color.fg_grey}'
                f'ID: {o.id} {symbol.sep} '
                f'NAME: {o.name} {symbol.sep} '
                f'PROJECT: {o.project_name} {symbol.sep} '
                f'INCLUDE EXTRACT: {include_extract}{color.reset}'
            )
            res = getattr(server.download, f'{object_type}')(o.id, include_extract=include_extract)
            color_print(f'{symbol.success}  {res}', fg='green')
        return f'Successfully downloaded all {object_type}s'

    # Gets the ID, name, and project from the object in Tableau Server
    obj = getattr(server.get, object_type)(object_id, object_name, project_name)
    object_id = obj.id or object_id
    object_name = obj.name or object_name
    project_name = obj.project_name or project_name

    if download:
        print(
            f'{color.fg_yellow}DOWNLOADING {object_type.upper()} {symbol.arrow_r} {color.fg_grey}'
            f'ID: {object_id} {symbol.sep} '
            f'NAME: {object_name} {symbol.sep} '
            f'PROJECT: {project_name} {symbol.sep} '
            f'INCLUDE EXTRACT: {include_extract}{color.reset}'
        )
        res: str = getattr(server.download, object_type)(object_id, include_extract=include_extract)
        color_print(f'{symbol.success}  {res}', fg='green')
    elif publish:
        print(
            f'{color.fg_yellow}PUBLISHING {object_type.upper()} {symbol.arrow_r} {color.fg_grey}'
            f'ID: {object_id} {symbol.sep} '
            f'NAME: {object_name} {symbol.sep} '
            f'PROJECT NAME: {project_name}{color.reset}'
        )
        res: Datasource | Workbook = getattr(server.publish, object_type)(
            file_path, object_id, object_name, project_name,
            connection=connection
        )
        color_print(f'{symbol.success}  {project_name} / {object_name}:', fg='green')
        color_print(f'  {symbol.arrow_r} {res.webpage_url}', fg='cyan')
        # Open URL to the published datasource in the browser
        webbrowser.open(res.webpage_url)
    elif embed_connection:
        print(
            f'{color.fg_yellow}EMBEDDING DATASOURCE CONNECTION CREDS {symbol.arrow_r} {color.fg_grey}'
            f'ID: {object_id} {symbol.sep} '
            f'NAME: {object_name} {symbol.sep} '
            f'PROJECT NAME: {project_name}{color.reset}'
        )
        res: Connection = server.embed_datasource_credentials(object_id, connection, args.conn_type)
        color_print(f'{symbol.success}  {res}', fg='green')
    elif refresh:
        print(
            f'{color.fg_yellow}REFRESHING {object_type.upper()} {symbol.arrow_r} {color.fg_grey}'
            f'ID: {object_id} {symbol.sep} '
            f'NAME: {object_name} {symbol.sep} '
            f'PROJECT NAME: {project_name}{color.reset}'
        )
        res: Job = getattr(server.refresh, object_type)(object_id)
        color_print(f'{symbol.success}  {res}', fg='green')



from requests import Session
from tableau_utilities.tableau_server.static import validate_response


class Base:
    """ Base functionality inherited by TableauServer class, and Core TableauServer classes """
    def __init__(self, parent):
        self.session: Session = parent.session
        self.user: str = parent.user
        self._pw: str = parent._pw
        self._personal_access_token_secret: str = parent._personal_access_token_secret
        self.personal_access_token_name: str = parent.personal_access_token_name
        self.host: str = parent.host
        self.site: str = parent.site
        self.api: float = parent.api
        self._auth_token = parent._auth_token
        self.url: str = parent.url
        self.get = parent.get if hasattr(parent, 'get') else None

    def _get(self, url, headers=None, **params):
        """ GET request for the Tableau REST API

        Args:
            url (str): URL endpoint for GET call
            headers (dict): GET call header

        Returns: The response content as a JSON dict
        """

        res = self.session.get(url, headers=headers, **params)
        return validate_response(res)

    def _post(self, url, json=None, headers=None, **params):
        """ POST request for the Tableau REST API

        Args:
            url (str): URL endpoint for POST call
            json (dict): The POST call JSON payload
            headers (dict): POST call header

        Returns: The response content as a JSON dict
        """
        res = self.session.post(url, json=json, headers=headers, **params)
        return validate_response(res)

    def _put(self, url, json=None, headers=None, **params):
        """ PUT request for the Tableau REST API

        Args:
            url (str): URL endpoint for PUT call
            json (dict): The PUT call JSON payload
            headers (dict): PUT call header

        Returns: The response content as a JSON dict
        """
        res = self.session.put(url, json=json, headers=headers, **params)
        return validate_response(res)

    def _delete(self, url, headers=None, **params):
        """ DELETE request for the Tableau REST API

        Args:
            url (str): URL endpoint for DELETE call
            headers (dict): DELETE call header

        Returns: The response content as a JSON dict
        """
        res = self.session.delete(url, headers=headers, **params)
        return validate_response(res)



import argparse
import os
import shutil
from argparse import RawTextHelpFormatter
import yaml
import pkg_resources

import tableau_utilities.tableau_server.tableau_server as ts

from tableau_utilities.general.config_column_persona import personas
from tableau_utilities.general.cli_styling import Color, Symbol, color_print
from tableau_utilities.scripts.gen_config import generate_config
from tableau_utilities.scripts.merge_config import merge_configs
from tableau_utilities.scripts.server_info import server_info
from tableau_utilities.scripts.server_operate import server_operate
from tableau_utilities.scripts.datasource import datasource
from tableau_utilities.scripts.csv_config import csv_config

__version__ = pkg_resources.require("tableau_utilities")[0].version

parser = argparse.ArgumentParser(
    prog='tableau_utilities',
    description='Tableau Utilities CLI:\n'
                '-Manage Tableau Server/Online\n'
                '-Manage configurations to edit datasource metadata',
    formatter_class=RawTextHelpFormatter
)
parser.add_argument('--version', action='version', version=f'%(prog)s {__version__}',
                    help='Print the current version of the CLI')
parser.add_argument('-d', '--debugging_logs', action='store_true',
                    help='Print detailed logging to the console to debug CLI')
subparsers = parser.add_subparsers(title="commands", dest="command", help='You must choose a command.', required=True)

# GROUP: Tableau Server Authentication
group_server_auth = parser.add_argument_group(
    'server_auth', 'Authentication used when interacting with Tableau Server'
)
group_server_auth.add_argument(
    '-s', '--server',
    help='Tableau Server URL. i.e. <server_address> in https://<server_address>.online.tableau.com'
)
group_server_auth.add_argument(
    '-sn', '--site_name',
    help='Site name. i.e. <site> in https://<server_address>.online.tableau.com/#/site/<site>'
)
group_server_auth.add_argument('--api_version', help='Tableau API version')
group_server_auth.add_argument('-u', '--user', help='The Tableau Server Username. Must pair with --password')
group_server_auth.add_argument('-p', '--password', help='The Tableau Server Password. Must pair with --user')
group_server_auth.add_argument('-ts', '--token_secret',
                               help='The Tableau Server Personal Access Token Secret. Must pair with --token_name')
group_server_auth.add_argument('-tn', '--token_name',
                               help='The Tableau Server Personal Access Token Name. Must pair with --token_secret')

# GROUP: SETTINGS YAML
group_settings_yaml = parser.add_argument_group('settings.yaml', 'Authentication with settings in a .yaml file')
group_settings_yaml.add_argument('--settings_path', default='settings.yaml',
                                 help='Path to your local settings.yaml file (See sample_settings.yaml)')

# GROUP: Output Directory
group_output_dir = parser.add_argument_group(
    'output_dir',
    'Manage the local directory where files will read from and write to'
)
group_output_dir.add_argument('-o', '--output_dir', default='tmp_tdsx_and_config',
                              help='Specifies the folder to write the datasource and configs to')
group_output_dir.add_argument('-c', '--clean_dir', action='store_true',
                              help='Deletes the directory, and all files within, before running')

# GROUP: File Information
group_file = parser.add_argument_group(
    'file', 'Args for commands that require specific information for a Tableau File; '
            'i.e. datasource, generate_config, or merge_config'
)
group_file.add_argument('-l', '--location', choices=['local', 'online'],
                        help='Specify the location of the Tableau File Object.')
group_file.add_argument('-i', '--id', help='The ID for the Tableau File Object in Cloud/Server.')
group_file.add_argument('-n', '--name',
                        help='The name of the Tableau File Object in Cloud/Server; Use with --project_name.')
group_file.add_argument('-pn', '--project_name',
                        help='The project name for the Tableau File Object in Cloud/Server; Use with --name.')
group_file.add_argument('-f', '--file_path', help='The path to the local Tableau File.')
group_file.add_argument('--definitions_csv',
                        help='Path to a CSV of datasource column definitions, to be added to a config. '
                             'Use with generate_config or merge_config')
group_file.add_argument('--include_extract', action='store_true',
                        help='Includes the extract in the download if specified. '
                             'This will make downloads take a long time for large extracts.')
group_file.add_argument('-tds', '--save_tds', action='store_true',
                        help='Saves the TDS for the datasource to view the raw xml')

# GROUP: CONNECTION
group_connection = parser.add_argument_group(
    'connection', 'Args which provide information for a Tableau Datasource connection.'
)
group_connection.add_argument('--conn_user', help='Connection Username')
group_connection.add_argument('--conn_pw', help='Connection Password')
group_connection.add_argument('--conn_type', help='Connection type; e.g. snowflake')
group_connection.add_argument('--conn_db', help='Connection Database')
group_connection.add_argument('--conn_schema', help='Connection Schema')
group_connection.add_argument('--conn_host', help='Connection Host (URL)')
group_connection.add_argument('--conn_role', help='Connection Role')
group_connection.add_argument('--conn_warehouse', help='Connection Warehouse')

# SERVER INFO
parser_server_info = subparsers.add_parser(
    'server_info',
    help='Retrieve and view information from Tableau Cloud/Server'
)
parser_server_info.add_argument('-lo', '--list_object', choices=['datasource', 'project', 'workbook'], required=True,
                                help='Specify the type of object for the information.')
parser_server_info.add_argument('-lf', '--list_format', default='names',
                                choices=['names', 'names_ids', 'names_projects', 'ids_names', 'full_df',
                                         'full_dictionary',
                                         'full_dictionary_pretty'],
                                help='Set the fields and format for the information.')
parser_server_info.add_argument('-lsf', '--list_sort_field', default='name',
                                help='Choose the field for sorting the information.')
parser_server_info.set_defaults(func=server_info)

# SERVER OPERATE
parser_server_operate = subparsers.add_parser(
    'server_operate', help='Download, publish, and refresh objects on Tableau Cloud/Server'
)
parser_server_operate.add_argument('--embed_connection', action='store_true',
                                   help='Specify to embed credentials for a Tableau Datasource Connection.'
                                        ' --conn_user and --conn_pw are required')
parser_server_operate.add_argument('--download', choices=['datasource', 'workbook'],
                                   help='Specify to download a Tableau object')
parser_server_operate.add_argument(
    '--publish', choices=['datasource', 'workbook'],
    help='Specify to publish a Tableau object. '
         '(Optional) Include --conn_user and --conn_pw to embed creds for the connection.'
)
parser_server_operate.add_argument('--refresh', choices=['datasource', 'workbook'],
                                   help='Specify to refresh a Tableau object')
parser_server_operate.add_argument('--all',  action='store_true', help='Download all workbooks or datasources')
parser_server_operate.set_defaults(func=server_operate)

# DATASOURCE
parser_datasource = subparsers.add_parser('datasource', help='View and edit metadata about the datasource')
parser_datasource.add_argument(
    '--enforce_connection', action='store_true',
    help='Provide to enforce the Datasource Connection attributes; See connection CLI group.'
)
parser_datasource.add_argument('--delete', choices=['folder', 'column'],
                               help='Deletes the specified object. The name of the object must be specified; '
                                    '--folder_name --column_name')
parser_datasource.add_argument('--list', choices=['folders', 'columns', 'metadata', 'connections'],
                               help='Lists the specified objects.')
parser_datasource.add_argument('--folder_name', help='The name of the folder. Required for --delete folder')
parser_datasource.add_argument('--column_name', help='The local name of the column. Required.')
parser_datasource.add_argument('--remote_name', help='The remote (SQL) name of the column.')
parser_datasource.add_argument('--caption', help='Short name/Alias for the column')
parser_datasource.add_argument('--title_case_caption', action='store_true',
                               help='Converts caption to title case. Applied after --caption')
parser_datasource.add_argument('--persona', choices=list(personas.keys()),
                               help='The datatype persona of the column. Required for adding a new column')
parser_datasource.add_argument('--desc', help='A Tableau column description')
parser_datasource.add_argument('--calculation', help='A Tableau calculation')
parser_datasource.set_defaults(func=datasource)

# GENERATE CONFIG
parser_config_gen = subparsers.add_parser(
    'generate_config', help='Generate configs to programmatically manage metadata in Tableau '
                            'datasources via Airflow.\nRequires File arguments.\n')
parser_config_gen.add_argument('--file_prefix', action='store_true',
                               help='Adds a prefix of the --name File arg to the output configs names.')
parser_config_gen.set_defaults(func=generate_config)

# CONFIG TO CSV
parser_config_csv = subparsers.add_parser(
    'csv_config', help='Write a config to a csv with 1 row per column per datasource')
parser_config_csv.add_argument('-cl', '--config_list', nargs='+', help='The list of paths to the configs')
parser_config_csv.set_defaults(func=csv_config)


# MERGE CONFIG
parser_config_merge = subparsers.add_parser(
    'merge_config', help='Merge a new config into the existing master config', formatter_class=RawTextHelpFormatter)
parser_config_merge.add_argument(
    '-mw', '--merge_with', choices=['config', 'csv', 'generate_merge_all'],
    help='config: merge an existing config with a new config\n'
         'csv: merge an existing config with a csv containing datasource column definitions\n'
         'generate_merge_all: Runs generate_config to create a column_config and a calc_config from a datasource. '
         'Merges both with existing configs in a target directory')
parser_config_merge.add_argument('-e', '--existing_config', help='The path to the base configuration file.')
parser_config_merge.add_argument('-a', '--additional_config',
                                 help='The path to the configuration to add to --existing_config.')
parser_config_merge.add_argument('-m', '--merged_config', default='merged_config',
                                 help='Name of the file to write the merged configuration file to. '
                                      'i.e. "my_config" will output to "my_config.json".')
parser_config_merge.add_argument('-td', '--target_directory', default='merged_config',
                                 help='The path containing the existing configs. '
                                      'Use with --merge_with generate_merge_all')
parser_config_merge.set_defaults(func=merge_configs)


def validate_args_server_operate(args):
    """ Validate that combinations of args are present """
    if (args.name and not args.project_name) or (args.project_name and not args.name):
        parser.error('--name and --project_name are required together')

    if args.publish and (args.name is None or args.project_name is None or args.file_path is None):
        parser.error('--publish requires: --name --project_name and --file_path ')

    if not (args.download or args.publish or args.refresh or args.embed_connection):
        parser.error('server_operate must be called with one of: '
                     '--download --publish --refresh --embed_connection')

    if len([i for i in [args.download, args.publish, args.refresh, args.embed_connection] if i]) > 1:
        parser.error('server_operate cannot be called with more than one of: '
                     '--download --publish --refresh --embed_connection')

    if args.embed_connection and (args.conn_user is None or args.conn_pw is None):
        parser.error('Both --conn_user and --conn_pw must be provided with --embed_connection')


def validate_args_id_name_project(args):
    """ Validate that combinations of args are present

    """

    if args.location is None:
        parser.error(f'The {args.command} command requires --location')

    if args.location == 'local' and not args.file_path:
        parser.error('--location local requires a --file_path')

    if args.location == 'online':
        if args.id:
            pass
        elif (args.name and not args.project_name) or (
                args.project_name and not args.name):
            parser.error('--name and --project_name are required together')
        elif not args.id and not args.name and not args.project_name:
            parser.error(
                '--location online requires either a --id or a --name and --project_name')


def validate_args_command_datasource(args):
    """ Validates args for the datasource command """
    if args.delete == 'folder' and not args.folder_name:
        parser.error(f'{args.command} --delete folder requires --folder_name')

    if args.delete == 'column' and not args.column_name:
        parser.error(f'{args.command} --delete column requires --column_name')


def validate_args_command_merge_config(args):
    if args.merge_with in ('config', 'csv') and args.existing_config is None:
        parser.error(f'--merge_with {args.merge_with} requires --existing_config')
    if args.merge_with == 'csv' and args.definitions_csv is None:
        parser.error(f'--merge_with {args.merge_with} requires --definitions_csv')
    if args.merge_with == 'config' and args.additional_config is None:
        parser.error(f'--merge_with {args.merge_with} requires --existing_config')
    if args.merge_with == 'generate_merge_all' and args.target_directory is None:
        parser.error(f'--merge_with {args.merge_with} requires --target_directory')


def tableau_authentication(args):
    """ Creates the Tableau server authentication from a variety of methods for passing in credentials """
    debug = args.debugging_logs
    yaml_path = args.settings_path
    color = Color()
    symbol = Symbol()

    if debug:
        title = f'{symbol.line * 29} Tableau Server Authentication {symbol.line * 29}'
        sub_title = ' Credentials are prioritized by: CLI Arguments > Settings YAML > Environment Variables '
        title_color = {'fg': 'green'}
        color_print(title, **title_color)
        print(f'{color.fg_black}{color.bg_yellow}{sub_title}{color.reset}')

    # Set CLI Argument credentials
    creds = {
        'user': args.user,
        'password': args.password,
        'api_version': args.api_version,
        'token_name': args.token_name,
        'token_secret': args.token_secret,
        'site': args.site_name,
        'server': args.server
    }

    if debug:
        for cred_name, cred_value in creds.items():
            if cred_value:
                print(f'  {symbol.arrow_r} Using CLI Argument cred: '
                      f'{cred_name} = {color.fg_cyan}{cred_value}{color.reset}')

    # Set Settings YAML file credentials
    if yaml_path and os.path.exists(yaml_path):
        with open(yaml_path, 'r') as f:
            yaml_creds = yaml.safe_load(f).get('tableau_login', {})
        if debug and not yaml_creds:
            color_print('YAML detected, but "tableau_login" is not defined', fg='yellow')
        for cred_name, cred_value in yaml_creds.items():
            if cred_value and cred_name in creds and not creds[cred_name]:
                creds[cred_name] = cred_value
                if debug:
                    print(f'  {symbol.arrow_r} Using Settings YAML cred: '
                          f'{cred_name} = {color.fg_cyan}{cred_value}{color.reset}')

    # Set Environment Variables credentials
    env_creds = {
        'token_name': os.getenv("TABLEAU_PERSONAL_ACCESS_TOKEN_NAME"),
        'token_secret': os.getenv("TABLEAU_PERSONAL_ACCESS_TOKEN_VALUE"),
        'site': os.getenv("TABLEAU_SITENAME"),
        'server': os.getenv("TABLEAU_SERVER_ADDRESS")
    }
    for cred_name, cred_value in env_creds.items():
        if cred_value and not creds[cred_name]:
            creds[cred_name] = cred_value
            if debug:
                print(f'  {symbol.arrow_r} Using Environment Variable cred: '
                      f'{cred_name} = {color.fg_cyan}{cred_value}{color.reset}')

    # Validate the combinations for authentication methods
    # If one, but not both, of token_name/token_secret or username/password are provided, throw an error
    if (creds['token_secret'] or creds['token_name']) and not (creds['token_name'] and creds['token_secret']):
        parser.error('--token_secret and --token_name are required together')
    if (creds['user'] or creds['password']) and not (creds['password'] and creds['user']):
        parser.error('--password and --user are required together')

    if debug:
        # Prints ending lines based on the title and title color printed above
        color_print(symbol.line * len(title), **title_color)
        print()  # new line

    # Create the server object and run the functions
    t = ts.TableauServer(
        personal_access_token_name=creds['token_name'],
        personal_access_token_secret=creds['token_secret'],
        user=creds['user'],
        password=creds['password'],
        site=creds['site'],
        host=f'https://{creds["server"]}.online.tableau.com',
        api_version=creds['api_version']
    )
    if debug:
        color_print(symbol.success, ' Connected to Tableau Server', **title_color)

    return t


def set_datasource_connection_args(args):
    """ Updates the connection group args, using a settings YAML file, and/or Environment Variables """
    debug = args.debugging_logs
    yaml_path = args.settings_path
    color = Color()
    symbol = Symbol()

    if debug:
        title = f'{symbol.line * 15} Setting Connection Group Arguments {symbol.line * 15}'
        sub_title = ' Credentials are prioritized by: CLI Arguments > Settings YAML > Environment Variables '
        title_color = {'fg': 'green'}
        color_print(title, **title_color)
        print(f'{color.fg_black}{color.bg_yellow}{sub_title}{color.reset}')

    # Set CLI Argument credentials
    creds = {
        'conn_user': args.conn_user,
        'conn_pw': args.conn_pw,
        'conn_type': args.conn_type,
        'conn_db': args.conn_db,
        'conn_schema': args.conn_schema,
        'conn_host': args.conn_host,
        'conn_role': args.conn_role,
        'conn_warehouse': args.conn_warehouse
    }
    if debug:
        for cred_name, cred_value in creds.items():
            if cred_value:
                print(f'  {symbol.arrow_r} Using CLI Argument cred: '
                      f'{cred_name} = {color.fg_cyan}{cred_value}{color.reset}')

    # Set Settings YAML file credentials
    if yaml_path and os.path.exists(yaml_path):
        with open(yaml_path, 'r') as f:
            yaml_creds = yaml.safe_load(f).get('datasource_connection', {})
        if debug and not yaml_creds:
            color_print('YAML detected, but "datasource_connection" is not defined', fg='yellow')
        for cred_name, cred_value in yaml_creds.items():
            if cred_value and cred_name in creds and not creds[cred_name]:
                creds[cred_name] = cred_value
                if debug:
                    print(f'  {symbol.arrow_r} Using Settings YAML cred: '
                          f'{cred_name} = {color.fg_cyan}{cred_value}{color.reset}')

    # Set Environment Variables credentials
    env_creds = {
        'conn_type': os.getenv('CONN_TYPE'),
        'conn_host': os.getenv('CONN_HOST'),
        'conn_user': os.getenv('CONN_USERNAME'),
        'conn_pw': os.getenv('CONN_PASSWORD'),
        'conn_db': os.getenv('CONN_DB'),
        'conn_schema': os.getenv('CONN_SCHEMA'),
        'conn_role': os.getenv('CONN_ROLE'),
        'conn_warehouse': os.getenv('CONN_WAREHOUSE')
    }
    for cred_name, cred_value in env_creds.items():
        if cred_value and not creds[cred_name]:
            creds[cred_name] = cred_value
            if debug:
                print(f'  {symbol.arrow_r} Using Environment Variable cred: '
                      f'{cred_name} = {color.fg_cyan}{cred_value}{color.reset}')

    if debug:
        # Prints ending lines based on the title and title color printed above
        color_print(symbol.line * len(title), **title_color)
        print()  # new line

    # Update Datasource Connection Group Args
    for arg, value in creds.items():
        setattr(args, arg, value)

    if debug:
        color_print(symbol.success, ' Set Datasource Connection Group Args', **title_color)


def main():
    args = parser.parse_args()

    # Set absolute path of the settings_path, if it exists and is not already absolute
    if not os.path.isabs(args.settings_path) and os.path.exists(args.settings_path):
        args.settings_path = os.path.abspath(args.settings_path)

    # Set args from connection group, from settings YAML / Environment Variables, when not provided in the command
    set_datasource_connection_args(args)

    # Validate the arguments
    if args.location == 'local' and args.file_path is None:
        parser.error('--location local requires --file_path')
    if args.command == 'server_operate':
        validate_args_server_operate(args)
    if args.command in ['generate_config', 'datasource']:
        validate_args_id_name_project(args)
    if args.command == 'datasource':
        validate_args_command_datasource(args)
    if args.command == 'merge_config':
        validate_args_command_merge_config(args)

    # Set/Reset the directory
    tmp_folder = args.output_dir
    if args.clean_dir:
        shutil.rmtree(tmp_folder, ignore_errors=True)

    os.makedirs(tmp_folder, exist_ok=True)
    os.chdir(tmp_folder)

    needs_tableau_server = (
        (args.command == 'generate_config' and args.location == 'online')
        or (args.command == 'merge_config' and args.location == 'online')
        or (args.command == 'datasource' and args.location == 'online')
        or (args.command == 'connection' and args.connection_operation == 'embed_user_pass')
        or args.command == 'server_info'
        or args.command == 'server_operate'
    )

    if needs_tableau_server:
        server = tableau_authentication(args)
        args.func(args, server)
    # Run functions that don't need the server
    else:
        args.func(args)


if __name__ == '__main__':
    main()



import pytest
import tableau_utilities as tu
import tableau_utilities.tableau_file.tableau_file_objects as tfo
import shutil
import os


EXTRACT_PATH = 'test_data_source.tdsx'
LIVE_PATH = 'test_live_data_source.tdsx'
ONE_FOLDER_PATH = 'one_folder.tdsx'
NO_FOLDER_PATH = 'no_folder.tdsx'

COLUMN = tfo.Column(
    name='FRIENDLY_NAME',
    caption='Friendly Name',
    datatype='string',
    type='ordinal',
    role='dimension',
    desc='Nice and friendly',
)
FOLDER = tfo.Folder(name='Friendly Name')


# cli.parser args
def test_cli_local_args():
    argv = [
        '-l', 'local',
        '-f', EXTRACT_PATH,
        '-tds',
        'datasource'
    ]
    args = tu.cli.parser.parse_args(argv)
    assert args.file_path == EXTRACT_PATH


# cli.parser args
def test_cli_server_info_args():
    argv = [
        '-tn', 'token_name',
        '-ts', 'token_secret',
        '-sn', 'nice_site',
        '-s', 'us-east-1',
        'server_info',
        '--list_object', 'datasource',
        '--list_format', 'ids_names',
        '--list_sort_field', 'id',
    ]
    args = tu.cli.parser.parse_args(argv)
    assert args.server == 'us-east-1'


# Column()
def test_column_attributes():
    column_attributes = {
        'name': '[FRIENDLY_CALC]',
        'caption': 'Friendly Calc',
        'datatype': 'integer',
        'type': 'ordinal',
        'role': 'dimension',
        'desc': 'Nice and friendly',
    }
    column1 = tfo.Column(calculation='count(1)', **column_attributes)
    column2 = tfo.Column(calculation={'@formula': 'count(1)', '@class': 'tableau'}, **column_attributes)
    for attr, value in column_attributes.items():
        assert value == getattr(column1, attr)
        if attr == 'calculation':
            assert 'count(1)' == getattr(column2, attr)
        else:
            assert value == getattr(column2, attr)


# TableauFileObjects().pop()
def test_tableau_file_objects_pop():
    column = tfo.Column(**{
        'name': '[FRIENDLY_CALC]',
        'caption': 'Friendly Calc',
        'datatype': 'integer',
        'type': 'ordinal',
        'role': 'dimension',
        'desc': 'Nice and friendly',
    })
    columns = tfo.TableauFileObjects([column])
    columns.pop(column)
    assert len(columns) == 0


# Datasource().DatasourceItems
def test_datasource_items():
    shutil.copyfile(f'resources/{EXTRACT_PATH}', EXTRACT_PATH)
    datasource = tu.Datasource(EXTRACT_PATH)
    os.remove(EXTRACT_PATH)
    assert datasource.columns is not None and datasource.columns != []
    assert datasource.folders_common is not None and datasource.folders_common != []
    assert datasource.aliases is not None and datasource.aliases != []
    assert datasource.connection is not None and datasource.connection != []
    assert datasource.extract is not None and datasource.extract != []


# Datasource().unzip()
def test_unzip_tableau_file():
    shutil.copyfile(f'resources/{EXTRACT_PATH}', EXTRACT_PATH)
    path = tu.Datasource(EXTRACT_PATH).unzip()
    with open(path) as f:
        content = f.read()
    os.remove(EXTRACT_PATH)
    os.remove(path)
    assert content is not None


# Datasource().save()
def test_datasource_save():
    shutil.copyfile(f'resources/{EXTRACT_PATH}', EXTRACT_PATH)
    datasource_before = tu.Datasource(EXTRACT_PATH)
    datasource_before.save()
    datasource_after = tu.Datasource(EXTRACT_PATH)
    os.remove(EXTRACT_PATH)
    for before, after in zip(datasource_before.sections(), datasource_after.sections()):
        assert before == after


# del Datasource().<section>
def test_delete_datasource_section():
    shutil.copyfile(f'resources/{EXTRACT_PATH}', EXTRACT_PATH)
    datasource_before = tu.Datasource(EXTRACT_PATH)
    extract_before = datasource_before.extract
    del datasource_before.extract
    datasource_before.save()
    datasource_after = tu.Datasource(EXTRACT_PATH)
    os.remove(EXTRACT_PATH)
    assert extract_before and not datasource_after.extract


# Datasource().columns.add()
def test_add_column():
    shutil.copyfile(f'resources/{EXTRACT_PATH}', EXTRACT_PATH)
    datasource = tu.Datasource(EXTRACT_PATH)
    os.remove(EXTRACT_PATH)
    datasource.columns.add(COLUMN)
    column = datasource.columns.get(COLUMN)
    assert column is not None


# Datasource().enforce_column()
def test_enforce_column():
    shutil.copyfile(f'resources/{EXTRACT_PATH}', EXTRACT_PATH)
    datasource = tu.Datasource(EXTRACT_PATH)
    os.remove(EXTRACT_PATH)
    col = tfo.Column(
        name='renamed_name',
        datatype='string',
        role='dimension',
        type='nominal',
        caption='Renamed Name'
    )
    datasource.enforce_column(col, folder_name='A New Folder', remote_name='NAME')
    column = datasource.columns.get(col)
    folder = datasource.folders_common.get('A New Folder')
    metadata = datasource.connection.metadata_records.get('NAME')
    assert column == col
    assert folder is not None
    assert metadata.local_name == '[renamed_name]'


# Datasource().columns.add()
def test_add_existing_column():
    shutil.copyfile(f'resources/{EXTRACT_PATH}', EXTRACT_PATH)
    datasource = tu.Datasource(EXTRACT_PATH)
    os.remove(EXTRACT_PATH)
    datasource.columns.add(COLUMN)
    datasource.columns.add(COLUMN)
    assert len([c for c in datasource.columns if c == COLUMN]) == 1


# Datasource().folders_common.folder.add()
def test_add_folder_tdsx_has_folder():
    shutil.copyfile(f'resources/{EXTRACT_PATH}', EXTRACT_PATH)
    datasource = tu.Datasource(EXTRACT_PATH)
    os.remove(EXTRACT_PATH)
    datasource.folders_common.add(FOLDER)
    found_folder = datasource.folders_common.get(FOLDER)
    assert found_folder is not None


# Datasource().folders_common.folder.add()
def test_add_folder_tdsx_does_not_have_folder():
    shutil.copyfile(f'resources/{NO_FOLDER_PATH}', NO_FOLDER_PATH)
    datasource = tu.Datasource(NO_FOLDER_PATH)
    os.remove(NO_FOLDER_PATH)
    datasource.folders_common.add(FOLDER)
    found_folder = datasource.folders_common.get(FOLDER)
    assert found_folder is not None


# Datasource().folders_common.folder.add()
def test_add_folder_tdsx_has_one_folder():
    shutil.copyfile(f'resources/{EXTRACT_PATH}', EXTRACT_PATH)
    datasource = tu.Datasource(EXTRACT_PATH)
    os.remove(EXTRACT_PATH)
    datasource.folders_common.add(FOLDER)
    found_folder = datasource.folders_common.get(FOLDER)
    assert found_folder is not None


# Datasource().folders_common.add()
def test_add_existing_folder():
    shutil.copyfile(f'resources/{EXTRACT_PATH}', EXTRACT_PATH)
    datasource = tu.Datasource(EXTRACT_PATH)
    os.remove(EXTRACT_PATH)
    datasource.folders_common.add(FOLDER)
    datasource.folders_common.add(FOLDER)
    assert len([f for f in datasource.folders_common if f == FOLDER]) == 1


# Datasource().folders_common.delete()
def test_delete_folder():
    shutil.copyfile(f'resources/{EXTRACT_PATH}', EXTRACT_PATH)
    datasource = tu.Datasource(EXTRACT_PATH)
    os.remove(EXTRACT_PATH)
    datasource.folders_common.add(FOLDER)
    datasource.folders_common.delete(FOLDER)
    found_folder = datasource.folders_common.get(FOLDER)
    assert not found_folder


# Datasource().columns.update()
def test_update_column():
    shutil.copyfile(f'resources/{EXTRACT_PATH}', EXTRACT_PATH)
    datasource = tu.Datasource(EXTRACT_PATH)
    os.remove(EXTRACT_PATH)
    column = tfo.Column(
        name='QUANTITY',
        caption='Quantity Renamed',
        datatype="integer",
        role="measure",
        type="quantitative"
    )
    datasource.columns.update(column)
    col = datasource.columns.get(column)
    assert col.caption == 'Quantity Renamed'


# Datasource().folders_common.get()
def test_get_folder():
    shutil.copyfile(f'resources/{EXTRACT_PATH}', EXTRACT_PATH)
    shutil.copyfile(f'resources/{LIVE_PATH}', LIVE_PATH)
    shutil.copyfile(f'resources/{ONE_FOLDER_PATH}', ONE_FOLDER_PATH)
    ds_extract = tu.Datasource(EXTRACT_PATH)
    ds_live = tu.Datasource(LIVE_PATH)
    ds_one_folder = tu.Datasource(ONE_FOLDER_PATH)
    os.remove(EXTRACT_PATH)
    os.remove(LIVE_PATH)
    os.remove(ONE_FOLDER_PATH)
    assert ds_extract.folders_common.get('tidy') is not None
    assert ds_live.folders_common.get('tidy') is None
    assert ds_one_folder.folders_common.get('neat') is not None


# Datasource().folders_common
def test_find_all_folders():
    shutil.copyfile(f'resources/{EXTRACT_PATH}', EXTRACT_PATH)
    shutil.copyfile(f'resources/{LIVE_PATH}', LIVE_PATH)
    shutil.copyfile(f'resources/{ONE_FOLDER_PATH}', ONE_FOLDER_PATH)
    ds_extract = tu.Datasource(EXTRACT_PATH)
    ds_live = tu.Datasource(LIVE_PATH)
    ds_one_folder = tu.Datasource(ONE_FOLDER_PATH)
    os.remove(EXTRACT_PATH)
    os.remove(LIVE_PATH)
    os.remove(ONE_FOLDER_PATH)
    # The new xml structure does not specify role
    _folder = ds_extract.folders_common
    _folder_live = ds_live.folders_common
    _folder_one_folder = ds_one_folder.folders_common
    assert _folder == ['neat', 'tidy']
    assert _folder_live == []
    assert _folder_one_folder == ['neat']


# Datasource().columns
def test_find_all_columns():
    shutil.copyfile(f'resources/{EXTRACT_PATH}', EXTRACT_PATH)
    shutil.copyfile(f'resources/{LIVE_PATH}', LIVE_PATH)
    ds_extract = tu.Datasource(EXTRACT_PATH)
    ds_live = tu.Datasource(LIVE_PATH)
    os.remove(EXTRACT_PATH)
    os.remove(LIVE_PATH)
    assert ds_extract.columns == [
        'CREATED_AT',
        'ID',
        'NAME',
        'Number of Records',
        'QUANTITY',
        '[__tableau_internal_object_id__].[Migrated Data]'
    ]
    assert ds_live.columns == [
        'Number of Records',
        '[__tableau_internal_object_id__].[Migrated Data]'
    ]


# Datasource().connection.relation.connection
def test_find_all_connections():
    shutil.copyfile(f'resources/{EXTRACT_PATH}', EXTRACT_PATH)
    shutil.copyfile(f'resources/{LIVE_PATH}', LIVE_PATH)
    ds_extract = tu.Datasource(EXTRACT_PATH)
    ds_live = tu.Datasource(LIVE_PATH)
    os.remove(EXTRACT_PATH)
    os.remove(LIVE_PATH)
    assert ds_extract.connection.named_connections == ['snowflake']
    assert ds_live.connection.named_connections == ['snowflake']


# get_connections()
def test_get_connection():
    shutil.copyfile(f'resources/{EXTRACT_PATH}', EXTRACT_PATH)
    shutil.copyfile(f'resources/{LIVE_PATH}', LIVE_PATH)
    ds_extract = tu.Datasource(EXTRACT_PATH)
    ds_live = tu.Datasource(LIVE_PATH)
    os.remove(EXTRACT_PATH)
    os.remove(LIVE_PATH)
    assert ds_extract.connection.named_connections.get('snowflake') is not None
    assert ds_live.connection.named_connections.get('snowflake') is not None


# Datasource().connection.update()
def test_update_connection():
    shutil.copyfile(f'resources/{EXTRACT_PATH}', EXTRACT_PATH)
    shutil.copyfile(f'resources/{LIVE_PATH}', LIVE_PATH)
    ds_extract = tu.Datasource(EXTRACT_PATH)
    ds_live = tu.Datasource(LIVE_PATH)
    os.remove(EXTRACT_PATH)
    os.remove(LIVE_PATH)
    connection = tfo.Connection(
        class_name='snowflake',
        dbname='FAKE_DB',
        schema='FAKE_SCHEMA',
        server='my_account_name.snowflakecomputing.com',
        service='FAKE_ROLE',
        username='FAKE_USER',
        warehouse='FAKE_WAREHOUSE'
    )
    ds_extract.connection.update(connection)
    ds_live.connection.update(connection)
    assert ds_extract.connection.named_connections['snowflake'].caption == connection.server
    assert ds_live.connection.named_connections['snowflake'].caption == connection.server
    assert ds_extract.connection['snowflake'].dbname == 'FAKE_DB'
    assert ds_live.connection['snowflake'].dbname == 'FAKE_DB'



import os
import yaml
import shutil
import xml.etree.ElementTree as ET
from pprint import pprint, isreadable
from time import time
from tableau_utilities import Datasource, TableauServer
from tableau_utilities import tableau_file_objects as tfo


if __name__ == '__main__':
    start = time()
    with open('../settings.yaml') as f:
        creds = yaml.safe_load(f)['tableau_login']
        personal_access_token_name = creds['token_name']
        personal_access_token_secret = creds['token_secret']
        site = creds['site']
        host = f'https://{creds["server"]}.online.tableau.com'
        api_version = creds['api_version']
    tmp_folder = 'tmp_downloads'
    # Cleanup lingering files, if there are any
    # shutil.rmtree(tmp_folder, ignore_errors=True)

    # Create a temp directory for testing
    os.makedirs(tmp_folder, exist_ok=True)
    os.chdir(tmp_folder)

    #  ### ### Testing here ### ###

    # datasource = Datasource('../resources/test_data_source.tdsx')
    # datasource.unzip()
    # datasource = Datasource('../resources/no_folder.tdsx')
    # datasource.unzip()
    # datasource = Datasource('test_data_source.tds')
    # datasource = Datasource('no_folder (local copy).tds')
    # folder = tfo.Folder('Test')
    # datasource.folders_common.add(folder)
    # datasource.save()
    # datasource = Datasource('test_data_source.tds')
    # datasource = Datasource('no_folder (local copy).tds')
    # pprint(datasource.folders_common)
    # datasource.save()
    # datasource.unzip()

    # Loop through the datasources
    # for d in datasources:
    # for path in os.listdir():
    #     if path != 'Jobs.tdsx':
    #         continue
    #     path = ts.download_datasource(datasource_id=d.id, include_extract=True)
    #     print('Downloaded:', path)
    #     datasource = Datasource(path)
    #     pprint(datasource.folders_common.folder)
    #     print(datasource.columns['job_id'].caption)
    #     datasource.columns['job_id'].caption = 'Job ID RENAMED'
    #     datasource._tree.write(path.replace('.tdsx',  '.tds'))
    #     datasource.save()
    #     datasource = Datasource(path)
    #     print(datasource.columns['job_id'].caption)
    #     column = tfo.Column(
    #         name='FRIENDLY_NAME',
    #         caption='Friendly Name',
    #         datatype='string',
    #         type='ordinal',
    #         role='dimension',
    #         desc='Nice and friendly',
    #     )
    #     datasource.enforce_column(column, remote_name='ORG_ID', folder_name='Org')
    #     print(datasource.folders.get('Org'))
    #     print(datasource.datasource_metadata.get('ORG_ID'))
    #     print(datasource.extract_metadata.get('ORG_ID'))
    #     print(datasource.datasource_mapping_cols.get(column.name))
    #     print(datasource.extract_mapping_cols.get(column.name))

    # ### ### ### ### ### ### ###

    # Cleanup lingering files, if there are any
    # os.chdir('..')
    # shutil.rmtree(tmp_folder)
    print(f'Script finished in {round(time() - start)} seconds')



import requests
from tableau_utilities.tableau_server.base import Base
from tableau_utilities.tableau_server.get import Get
from tableau_utilities.tableau_server.create import Create
from tableau_utilities.tableau_server.download import Download
from tableau_utilities.tableau_server.publish import Publish
from tableau_utilities.tableau_server.refresh import Refresh
from tableau_utilities.tableau_server.update import Update
from tableau_utilities.tableau_server.static import TableauConnectionError


class TableauServer(Base):
    """ Connects and interacts with Tableau Online/Server, via the REST API. """

    def __init__(
            self,
            host: str,
            site: str,
            user: str = None,
            password: str = None,
            personal_access_token_secret: str = None,
            personal_access_token_name: str = None,
            api_version: float = None
    ):
        """ To sign in to Tableau a user needs either a username & password or token secret & token name

        Args:
            host: Tableau server address
            user: The username to sign in to Tableau Online with
            password: The password to sign in to Tableau Online with
            personal_access_token_name: The name of the personal access token used
            personal_access_token_secret: The secret of the personal access token used
            site: The Tableau Online site id
            api_version: The Tableau REST API version
        """
        self.user = user
        self._pw = password
        self._personal_access_token_secret = personal_access_token_secret
        self.personal_access_token_name = personal_access_token_name
        self.host = host
        self.site = site
        self.api = api_version or 3.18
        # Set by class
        self._auth_token = None
        self.url: str = None
        # Create a session on initialization
        self.session = requests.session()
        self.session.headers.update({'accept': 'application/json', 'content-type': 'application/json'})
        super().__init__(self)
        # Sign in on initialization
        self.__sign_in()
        # Assign core functionality
        self.get: Get = Get(self)
        self.create: Create = Create(self)
        self.download: Download = Download(self)
        self.publish: Publish = Publish(self)
        self.refresh: Refresh = Refresh(self)
        self.update: Update = Update(self)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Sign out upon class deconstruction
        print('Signing out of Tableau Server')
        self.sign_out()

    def __sign_in(self):
        """
            To sign in to Tableau a user needs either a username & password or token secret & token name
            Signs in to the server with credentials from the specified connection.
            Sets the auth_token, site_id, and url common prefix

        """
        url = f"{self.host}/api/{self.api}/auth/signin"
        if self._personal_access_token_secret and self.personal_access_token_name:
            body = {"credentials": {"personalAccessTokenSecret": self._personal_access_token_secret,
                                    "personalAccessTokenName": self.personal_access_token_name,
                                    "site": {"contentUrl": self.site}}}
        elif self.user and self._pw:
            body = {"credentials": {"name": self.user, "password": self._pw, "site": {"contentUrl": self.site}}}
        else:
            raise TableauConnectionError(
                'Please provide either user and password, or token_secret and token_name'
            )

        res = self._post(url, json=body).get('credentials', {})
        # Set auth token and site ID attributes on sign in
        self.session.headers.update({'x-tableau-auth': res.get('token')})
        self.site = res.get('site', {}).get('id')
        self.url = f"{self.host}/api/{self.api}/sites/{self.site}"

    def sign_out(self):
        """ Destroys the active session and invalidates authentication token. """
        self._post(url=f"{self.host}/api/{self.api}/auth/signout")
        self.session.close()

    def embed_datasource_credentials(self, datasource_id, credentials, connection_type):
        """ Embed the given credentials for all connections of a datasource of the given connection type.
            Only embeds Username and Password credentials.

        Args:
            datasource_id (str): The ID of the datasource
            connection_type (str): Type of connection you want to embed creds for, i.e. snowflake
            credentials (dict): The credentials dict to embed
                i.e. {'username': 'user', 'password': 'password'}
        """
        for cred in ['username', 'password']:
            if not credentials.get(cred):
                raise TableauConnectionError(f'Missing required credential: {cred}')

        for c in self.get.datasource_connections(datasource_id):
            if c.type.lower() == connection_type.lower():
                c.user_name = credentials['username']
                c.password = credentials['password']
                c.embed_password = True
                response = self.update.datasource_connection(datasource_id, c)
                return response
